(* collecting all knowledge for LinearEquations
   created by: rlang 
         date: 02.10
   changed by: rlang
   last change by: rlang
             date: 02.11.04
*)

(* remove_thy"LinEq";
   use_thy"knowledge/Isac";

   use_thy"knowledge/LinEq";

   use"ROOT.ML";
   cd"knowledge";
*)

"******* LinEq.ML begin *******";

(*-------------------- theory -------------------------------------------------*)
theory' := overwritel (!theory', [("LinEq.thy",LinEq.thy)]);

val lineq_prls = (*3.10.02:just the following order due to subterm evaluation*)
  append_rls "lineq_prls" e_rls 
	     [Calc ("op =",eval_equal "#equal_"),
	      Calc ("Tools.matches",eval_matches ""),
	      Calc ("Tools.lhs"    ,eval_lhs ""),
	      Calc ("Poly.has'_degree'_in",eval_has_degree_in ""),
 	      Calc ("Poly.is'_polyrat'_in",eval_is_polyrat_in ""),
	      Calc ("Atools.occurs'_in",eval_occurs_in ""),    
	      Calc ("Atools.ident",eval_ident "#ident_")
	      ];


(*-------------- rules -------------------------------------------------------*)
(* ----- erls ----- *)
val lineq_erls = 
    merge_rls "lineq_erls" poly_erls
    (append_rls "op_pred" e_rls
               [Calc ("op -",eval_binop "#subtract_"),
                Calc ("Poly.is'_polyrat'_in",eval_is_polyrat_in "")
		(*		
		 Don't use
		 Calc ("HOL.divide", eval_cancel "#divide_"),
		 Calc ("Atools.pow" ,eval_binop "#power_"),
		 *)
		 ]);

(*isolate the bound variable in an linear equation; 'bdv' is a meta-constant*)
val LinEq_simplify = 
Rls {id = "LinEq_simplify", preconds = [],
     rew_ord = ("e_rew_ord",e_rew_ord),
     erls = e_rls,
     srls = Erls,
     calc = [],
     asm_thm = [],
     rules = [
	      Thm("lin_isolate_add1",num_str lin_isolate_add1), 
	      (* a+bx=0 -> bx=-a *)
	      Thm("lin_isolate_add2",num_str lin_isolate_add2), 
	      (* a+ x=0 ->  x=-a *)
	      Thm("lin_isolate_div",num_str lin_isolate_div)    
	      (*   bx=c -> x=c/b *)  
	      ],
     scr = Script ((term_of o the o (parse thy)) "empty_script")
     }:rls;
ruleset' := overwritel (!ruleset',
			[("LinEq_simplify",LinEq_simplify),
			 ("lineq_erls",lineq_erls)(*FIXXXME:del with rls.rls'*)
			 ]);
val poly_simplify = 
  Rls {id = "poly_simplify", preconds = [], 
       rew_ord = ("termlessI",termlessI), 
       erls = lineq_erls, 
       srls = Erls, 
       calc = [], 
       asm_thm = [],
       rules = [
		Thm  ("real_assoc_1",num_str real_assoc_1),
		Calc ("op +",eval_binop "#add_"),
		Calc ("op -",eval_binop "#sub_"),
		Calc ("op *",eval_binop "#mult_"),
		(*  Dont use  
		 Calc ("HOL.divide", eval_cancel "#divide_"),		
		 Calc ("Root.sqrt",eval_sqrt "#sqrt_"),
		 *)
		Calc ("Atools.pow" ,eval_binop "#power_")
		],
       scr = Script ((term_of o the o (parse thy)) "empty_script")
       }:rls;
ruleset' := overwritel (!ruleset',
			  [("poly_simplify",poly_simplify)]);

(*----------------------------- problem types --------------------------------*)
(* 
show_ptyps(); 
(get_pbt ["linear","univariate","equation"]);
*)
(* ---------linear----------- *)
store_pbt
 (prep_pbt LinEq.thy
 (["linear","univariate","equation"],
  [("#Given" ,["equality e_","solveFor v_"]),
   ("#Where" ,["False",
               "Not( (lhs e_) is_polyrat_in v_)",
	       "( (lhs e_)  has_degree_in v_)=1"]),
   ("#Find"  ,["solutions v_i_"]) 
  ],
  lineq_prls, None,
  [("LinEq.thy","solve_lin_equation")]));

(*-------------- methods-------------------------------------------------------*)
(* ansprechen mit "LinEq.thy","solve_univar_equation" *)
methods:= overwritel (!methods,
[(* ----- 07.10.02---- *)
prep_met
 (("LinEq.thy","solve_lin_equation"),
   [("#Given" ,["equality e_","solveFor v_"]),
    ("#Where" ,["Not( (lhs e_) is_polyrat_in v_)",
                "( (lhs e_)  has_degree_in v_)=1"]),
    ("#Find"  ,["solutions v_i_"])
   ],
   {rew_ord'="termlessI",
    rls'=lineq_erls,
    srls=e_rls,
    prls=lineq_prls,
    calc=[],
    asm_rls=["LinEq_simplify"],
    asm_thm=[]},
    "Script Solve_lin_equation (e_::bool) (v_::real) =                   \
    \(let e_ =((Try         (Rewrite     all_left            False)) @@  \ 
    \          (Try (Repeat (Rewrite     makex1_x           False))) @@  \ 
    \          (Try         (Rewrite_Set expand_binoms       False)) @@  \ 
    \          (Try         (Rewrite_Set make_polynomial     False)) @@  \
    \          (Try (Repeat (Rewrite_Set poly_simplify      False)))) e_;\
    \     e_ = ((Try (Rewrite_Set_Inst [(bdv,v_::real)]                  \
    \                                          LinEq_simplify True)) @@  \
    \            (Repeat(Try (Rewrite_Set poly_simplify     False)))) e_ \
    \ in ((Or_to_List e_)::bool list))"
 )
]);
"******* LinEq.ML end *******";
