(* (c) by Richard Lang
   theory collecting all knowledge for RootEquations
   created by: rlang 
         date: 02.09
   changed by: rlang
   last change by: rlang
             date: 02.11.14
*)

(* use"knowledge/RootEq.ML";
   use"RootEq.ML";
 
   use"ROOT.ML";
   cd"knowledge";
 
   remove_thy"RootEq";
   use_thy"Isac";
   *)
"******* RootEq.ML begin *******";

theory' := overwritel (!theory', [("RootEq.thy",RootEq.thy)]);
(*-------------------------functions---------------------*)
(* true if bdv is under sqrt of a Equation*)
fun is_rootequation_in t v = 
    let 
	fun coeff_in c v = v mem (vars c);
   	fun findroot (_ $ _ $ _ $ _) v = raise error("is_rootequation_in:")
	  (* at the moment there is no term like this, but ....*)
	  | findroot (t as (Const ("Root.nroot",_) $ _ $ t3)) v = coeff_in t3 v
	  | findroot (_ $ t2 $ t3) v = (findroot t2 v) orelse (findroot t3 v)
	  | findroot (t as (Const ("Root.sqrt",_) $ t2)) v = coeff_in t2 v
	  | findroot (_ $ t2) v = (findroot t2 v)
	  | findroot _ _ = false;
     in
	findroot t v
    end;

 fun is_sqrtequation_in t v = 
    let 
	fun coeff_in c v = v mem (vars c);
   	fun findsqrt (_ $ _ $ _ $ _) v = raise error("is_sqrteqation_in:")
	  (* at the moment there is no term like this, but ....*)
	  | findsqrt (_ $ t1 $ t2) v = (findsqrt t1 v) orelse (findsqrt t2 v)
	  | findsqrt (t as (Const ("Root.sqrt",_) $ a)) v = coeff_in a v
	  | findsqrt (_ $ t1) v = (findsqrt t1 v)
	  | findsqrt _ _ = false;
     in
	findsqrt t v
    end;
    
fun eval_is_rootequation_in _ _ (p as (Const ("RootEq.is'_rootequation'_in",_) $ t $ v)) _  =
    if is_rootequation_in t v then 
	Some ((term2str p) ^ " = True",
	      Trueprop $ (mk_equality (p, HOLogic.true_const)))
    else Some ((term2str p) ^ " = True",
	       Trueprop $ (mk_equality (p, HOLogic.false_const)))
  | eval_is_rootequation_in _ _ _ _ = ((*writeln"### nichts matcht";*) None);

fun eval_is_sqrtequation_in _ _ (p as (Const ("RootEq.is'_sqrtequation'_in",_) $ t $ v)) _  =
    if is_sqrtequation_in t v then 
	Some ((term2str p) ^ " = True",
	      Trueprop $ (mk_equality (p, HOLogic.true_const)))
    else Some ((term2str p) ^ " = True",
	       Trueprop $ (mk_equality (p, HOLogic.false_const)))
  | eval_is_sqrtequation_in _ _ _ _ = ((*writeln"### nichts matcht";*) None);

(*-------------------------rulse-------------------------*)
val rooteq_prls = (*15.10.02:just the following order due to subterm evaluation*)
  append_rls "rooteq_prls" e_rls 
	     [Calc ("Atools.ident",eval_ident "#ident_"),
	      Calc ("Tools.matches",eval_matches ""),
	      Calc ("Tools.lhs"    ,eval_lhs ""),
	      Calc ("RootEq.is'_sqrtequation'_in",eval_is_sqrtequation_in ""),
	      Calc ("RootEq.is'_rootequation'_in",eval_is_rootequation_in ""),
	      Calc ("op =",eval_equal "#equal_"),
	      Thm ("not_true",num_str not_true),
	      Thm ("not_false",num_str not_false),
	      Thm ("and_true",num_str and_true),
	      Thm ("and_false",num_str and_false),
	      Thm ("or_true",num_str or_true),
	      Thm ("or_false",num_str or_false)
	      (* Thm ("and_commute",num_str and_commute),
	       Thm ("or_commute",num_str or_commute)
	       *)
	      ];

val rooteq_erls = 
    merge_rls "rooteq_erls" root_erls
      (append_rls "real_divide_divide2_eq" e_rls
		[Thm ("real_divide_divide2_eq",num_str real_divide_divide2_eq)
		 ]);

val rooteq_srls = 
     append_rls "rooteq_srls" e_rls
		[Calc ("RootEq.is'_sqrtequation'_in",
		       eval_is_sqrtequation_in ""),
                 Calc ("RootEq.is'_rootequation'_in",eval_is_rootequation_in "")
		 ];

ruleset' := overwritel (!ruleset',
			[("rooteq_erls",rooteq_erls),(*FIXXXME:del with rls.rls'*)
			 ("rooteq_srls",rooteq_srls)
                         ]);

(*isolate the bound variable in an sqrt equation; 'bdv' is a meta-constant*)
 val sqrt_isolate = 
  Rls {id = "sqrt_isolate", preconds = [], rew_ord = ("termlessI",termlessI), 
       erls = rooteq_erls, srls = Erls, calc = [], 
       asm_thm = [("sqrt_square_1",""),("sqrt_square_equation_left_1",""),
                  ("sqrt_square_equation_left_2",""),("sqrt_square_equation_left_3",""),
                  ("sqrt_square_equation_left_4",""),("sqrt_square_equation_right_1",""),
                  ("sqrt_square_equation_right_2",""),("sqrt_square_equation_right_3",""),
                  ("sqrt_square_equation_right_4","")],
       rules = [
	      Thm("sqrt_square_1",num_str sqrt_square_1),                            (* (sqrt a)^^^2 -> a *)
	      Thm("sqrt_square_2",num_str sqrt_square_2),                            (* sqrt (a^^^2) -> a *)
	      Thm("sqrt_times_root_1",num_str sqrt_times_root_1),            (* sqrt a sqrt b -> sqrt(ab) *)
	      Thm("sqrt_times_root_2",num_str sqrt_times_root_2),        (* a sqrt b sqrt c -> a sqrt(bc) *)
              Thm("sqrt_square_equation_both_1",num_str sqrt_square_equation_both_1),
              (* (sqrt a + sqrt b  = sqrt c + sqrt d) -> (a+2*sqrt(a)*sqrt(b)+b) = c+2*sqrt(c)*sqrt(d)+d) *)
              Thm("sqrt_square_equation_both_2",num_str sqrt_square_equation_both_2),
              (* (sqrt a - sqrt b  = sqrt c + sqrt d) -> (a-2*sqrt(a)*sqrt(b)+b) = c+2*sqrt(c)*sqrt(d)+d) *)
              Thm("sqrt_square_equation_both_3",num_str sqrt_square_equation_both_3),
              (* (sqrt a + sqrt b  = sqrt c - sqrt d) -> (a+2*sqrt(a)*sqrt(b)+b) = c-2*sqrt(c)*sqrt(d)+d) *)
              Thm("sqrt_square_equation_both_4",num_str sqrt_square_equation_both_4),
              (* (sqrt a - sqrt b  = sqrt c - sqrt d) -> (a-2*sqrt(a)*sqrt(b)+b) = c-2*sqrt(c)*sqrt(d)+d) *)
	      Thm("sqrt_isolate_l_add1",num_str sqrt_isolate_l_add1), (* a+b*sqrt(x)=d -> b*sqrt(x) = d-a *)
	      Thm("sqrt_isolate_l_add2",num_str sqrt_isolate_l_add2), (* a+  sqrt(x)=d ->   sqrt(x) = d-a *)
	      Thm("sqrt_isolate_l_add3",num_str sqrt_isolate_l_add3), (* a+b*c/sqrt(x)=d->b*c/sqrt(x)=d-a *)
	      Thm("sqrt_isolate_l_add4",num_str sqrt_isolate_l_add4), (* a+c/sqrt(x)=d -> c/sqrt(x) = d-a *)
	      (*Thm("sqrt_isolate_l_div",num_str sqrt_isolate_l_div),*)      (* b*sqrt(x) = d sqrt(x) d/b *)
	      Thm("sqrt_isolate_r_add1",num_str sqrt_isolate_r_add1),  (* a= d+e*sqrt(x) -> a-d=e*sqrt(x) *)
	      Thm("sqrt_isolate_r_add2",num_str sqrt_isolate_r_add2),  (* a= d+  sqrt(x) -> a-d=  sqrt(x) *)
	      Thm("sqrt_isolate_r_add3",num_str sqrt_isolate_r_add3),  (* a=d+e*g/sqrt(x)->a-d=e*g/sqrt(x)*)
	      Thm("sqrt_isolate_r_add4",num_str sqrt_isolate_r_add4),  (* a= d+g/sqrt(x) -> a-d=g/sqrt(x) *)
	      (*Thm("sqrt_isolate_r_div",num_str sqrt_isolate_r_div),*)   (* a=e*sqrt(x) -> a/e = sqrt(x) *)
	      Thm("sqrt_square_equation_left_1",num_str sqrt_square_equation_left_1),   
	      (* sqrt(x)=b -> x=b^2 *)
	      Thm("sqrt_square_equation_left_2",num_str sqrt_square_equation_left_2),   
	      (* c*sqrt(x)=b -> c^2*x=b^2 *)
	      Thm("sqrt_square_equation_left_3",num_str sqrt_square_equation_left_3),   
	      (* c/sqrt(x)=b -> c^2/x=b^2 *)
	      Thm("sqrt_square_equation_left_4",num_str sqrt_square_equation_left_4),   
	      (* c*d/sqrt(x)=b -> c^2*d^2/x=b^2 *)
	      Thm("sqrt_square_equation_right_1",num_str sqrt_square_equation_right_1),   
	      (* a=sqrt(x) ->a^2=x *)
	      Thm("sqrt_square_equation_right_2",num_str sqrt_square_equation_right_2),   
	      (* a=c*sqrt(x) ->a^2=c^2*x *)
	      Thm("sqrt_square_equation_right_3",num_str sqrt_square_equation_right_3),   
	      (* a=c/sqrt(x) ->a^2=c^2/x *)
	      Thm("sqrt_square_equation_right_4",num_str sqrt_square_equation_right_4)   
	      (* a=c*d/sqrt(x) ->a^2=c^2*d^2/x *)
	      ],
	 scr = Script ((term_of o the o (parse thy)) "empty_script")
         }:rls;
ruleset' := overwritel (!ruleset',
			[("sqrt_isolate",sqrt_isolate)
			 ]);
(* -- left 28.08.02--*)
(*isolate the bound variable in an sqrt left equation; 'bdv' is a meta-constant*)
 val l_sqrt_isolate = 
     Rls {id = "l_sqrt_isolate", preconds = [], 
	  rew_ord = ("termlessI",termlessI), 
     erls = rooteq_erls, srls = Erls, calc = [], asm_thm = [],
     rules = [
	      Thm("sqrt_square_1",num_str sqrt_square_1),                            (* (sqrt a)^^^2 -> a *)
	      Thm("sqrt_square_2",num_str sqrt_square_2),                            (* sqrt (a^^^2) -> a *)
	      Thm("sqrt_times_root_1",num_str sqrt_times_root_1),            (* sqrt a sqrt b -> sqrt(ab) *)
	      Thm("sqrt_times_root_2",num_str sqrt_times_root_2),        (* a sqrt b sqrt c -> a sqrt(bc) *)
	      Thm("sqrt_isolate_l_add1",num_str sqrt_isolate_l_add1), (* a+b*sqrt(x)=d -> b*sqrt(x) = d-a *)
	      Thm("sqrt_isolate_l_add2",num_str sqrt_isolate_l_add2), (* a+  sqrt(x)=d ->   sqrt(x) = d-a *)
	      Thm("sqrt_isolate_l_add3",num_str sqrt_isolate_l_add3), (* a+b*c/sqrt(x)=d->b*c/sqrt(x)=d-a *)
	      Thm("sqrt_isolate_l_add4",num_str sqrt_isolate_l_add4), (* a+c/sqrt(x)=d -> c/sqrt(x) = d-a *)
	      (*Thm("sqrt_isolate_l_div",num_str sqrt_isolate_l_div),*)      (* b*sqrt(x) = d sqrt(x) d/b *)
	      Thm("sqrt_square_equation_left_1",num_str sqrt_square_equation_left_1),
	      (* sqrt(x)=b -> x=b^2 *)
	      Thm("sqrt_square_equation_left_2",num_str sqrt_square_equation_left_2),
	      (* a*sqrt(x)=b -> a^2*x=b^2*)
	      Thm("sqrt_square_equation_left_3",num_str sqrt_square_equation_left_3),   
	      (* c/sqrt(x)=b -> c^2/x=b^2 *)
	      Thm("sqrt_square_equation_left_4",num_str sqrt_square_equation_left_4)   
	      (* c*d/sqrt(x)=b -> c^2*d^2/x=b^2 *)
	      ],
	 scr = Script ((term_of o the o (parse thy)) "empty_script")
         }:rls;
ruleset' := overwritel (!ruleset',
			[("l_sqrt_isolate",l_sqrt_isolate)
			 ]);

(* -- right 28.8.02--*)
(*isolate the bound variable in an sqrt right equation; 'bdv' is a meta-constant*)
 val r_sqrt_isolate = 
     Rls {id = "r_sqrt_isolate", preconds = [], 
	  rew_ord = ("termlessI",termlessI), 
     erls = rooteq_erls, srls = Erls, calc = [], asm_thm = [],
     rules = [
	      Thm("sqrt_square_1",num_str sqrt_square_1),                           (* (sqrt a)^^^2 -> a *)
	      Thm("sqrt_square_2",num_str sqrt_square_2),                           (* sqrt (a^^^2) -> a *)
	      Thm("sqrt_times_root_1",num_str sqrt_times_root_1),           (* sqrt a sqrt b -> sqrt(ab) *)
	      Thm("sqrt_times_root_2",num_str sqrt_times_root_2),       (* a sqrt b sqrt c -> a sqrt(bc) *)
	      Thm("sqrt_isolate_r_add1",num_str sqrt_isolate_r_add1), (* a= d+e*sqrt(x) -> a-d=e*sqrt(x) *)
	      Thm("sqrt_isolate_r_add2",num_str sqrt_isolate_r_add2), (* a= d+  sqrt(x) -> a-d=  sqrt(x) *)
	      Thm("sqrt_isolate_r_add3",num_str sqrt_isolate_r_add3),  (* a=d+e*g/sqrt(x)->a-d=e*g/sqrt(x)*)
	      Thm("sqrt_isolate_r_add4",num_str sqrt_isolate_r_add4),  (* a= d+g/sqrt(x) -> a-d=g/sqrt(x) *)
	      (*Thm("sqrt_isolate_r_div",num_str sqrt_isolate_r_div),*)  (* a=e*sqrt(x) -> a/e = sqrt(x) *)
	      Thm("sqrt_square_equation_right_1",num_str sqrt_square_equation_right_1),
	      (* a=sqrt(x) ->a^2=x *)
	      Thm("sqrt_square_equation_right_2",num_str sqrt_square_equation_right_2),
	      (* a=c*sqrt(x) ->a^2=c^2*x *)
	      Thm("sqrt_square_equation_right_3",num_str sqrt_square_equation_right_3),   
	      (* a=c/sqrt(x) ->a^2=c^2/x *)
	      Thm("sqrt_square_equation_right_4",num_str sqrt_square_equation_right_4)   
	      (* a=c*d/sqrt(x) ->a^2=c^2*d^2/x *)
	      ],
	 scr = Script ((term_of o the o (parse thy)) "empty_script")
         }:rls;
ruleset' := overwritel (!ruleset',
			[("r_sqrt_isolate",r_sqrt_isolate)
			 ]);

val rooteq_simplify = 
  Rls {id = "rooteq_simplify", 
       preconds = [], rew_ord = ("termlessI",termlessI), 
      erls = rooteq_erls, srls = Erls, calc = [], asm_thm = [],
    rules = [Thm  ("real_assoc_1",num_str real_assoc_1),                             (* a+(b+c) = a+b+c *)
             Thm  ("real_assoc_2",num_str real_assoc_2),                             (* a*(b*c) = a*b*c *)
	     Calc ("op +",eval_binop "#add_"),
	     Calc ("op -",eval_binop "#sub_"),
	     Calc ("op *",eval_binop "#mult_"),
	     Calc ("HOL.divide", eval_cancel "#divide_"),
	     Calc ("Root.sqrt",eval_sqrt "#sqrt_"),
	     Calc ("Atools.pow" ,eval_binop "#power_"),
	     Thm("real_plus_binom_pow2",num_str real_plus_binom_pow2),
	     Thm("real_minus_binom_pow2",num_str real_minus_binom_pow2),
	     Thm("realpow_mul",num_str realpow_mul),    (* (a * b)^n = a^n * b^n*)
	     Thm("sqrt_times_root_1",num_str sqrt_times_root_1),         (* sqrt b * sqrt c = sqrt(b*c) *)
	     Thm("sqrt_times_root_2",num_str sqrt_times_root_2), (* a * sqrt a * sqrt b = a * sqrt(a*b) *)
	     Thm("sqrt_square_2",num_str sqrt_square_2),                            (* sqrt (a^^^2) = a *)
	     Thm("sqrt_square_1",num_str sqrt_square_1)                             (* sqrt a ^^^ 2 = a *)
	     ],
    scr = Script ((term_of o the o (parse thy)) "empty_script")
    }:rls;
ruleset' := overwritel (!ruleset',
			[("rooteq_simplify",rooteq_simplify)
			 ]);

(*-------------------------Problem-----------------------*)
(*
(get_pbt ["root","univariate","equation"]);
show_ptyps(); 
*)
store_pbt
 (prep_pbt RootEq.thy
 (["root","univariate","equation"],
  [("#Given" ,["equality e_","solveFor v_"]),
   ("#Where" ,["(lhs e_) is_rootequation_in  (v_::real) | \
	       \(rhs e_) is_rootequation_in  (v_::real)"]),
   ("#Find"  ,["solutions v_i_"]) 
  ],
  rooteq_prls, None,
  []));
(* ---------sqrt----------- *)
store_pbt
 (prep_pbt RootEq.thy
 (["sq","root","univariate","equation"],
  [("#Given" ,["equality e_","solveFor v_"]),
   ("#Where" ,["(lhs e_) is_sqrtequation_in (v_::real) | \
	       \(rhs e_) is_sqrtequation_in (v_::real)"]),
   ("#Find"  ,["solutions v_i_"]) 
  ],
  rooteq_prls,  None,
  [("RootEq.thy","solve_sq_root_equation")]));

store_pbt
 (prep_pbt RootEq.thy
 (["normalize","root","univariate","equation"],
  [("#Given" ,["equality e_","solveFor v_"]),
   ("#Where" ,["(rhs e_) is_sqrtequation_in (v_::real) | \
               \(lhs e_) is_sqrtequation_in (v_::real)",
	       "Not(matches ( ?a = ?b + ?c*sqrt(?d)) e_) &\
	       \Not(matches ( ?a = ?b +    sqrt(?d)) e_) &\
	       \Not(matches ( ?a =      ?c*sqrt(?d)) e_) &\
	       \Not(matches ( ?a =         sqrt(?d)) e_)&\
	       \Not(matches ( ?a + b*sqrt(?c) = ?d) e_) &\
	       \Not(matches ( ?a +   sqrt(?c) = ?d) e_) &\
	       \Not(matches (      b*sqrt(?c) = ?d) e_) &\
	       \Not(matches (        sqrt(?c) = ?d) e_)"]),
   ("#Find"  ,["solutions v_i_"]) 
  ],
  rooteq_prls,  None,
  [("RootEq.thy","norm_sq_root_equation")]));

(*-------------------------methods-----------------------*)
(* ---- root 20.8.02 ---*)
methods:= overwritel (!methods,
[ 
(*-- normalize 20.10.02 --*)
prep_met
 (("RootEq.thy","norm_sq_root_equation"),
   [("#Given" ,["equality e_","solveFor v_"]),
    ("#Where" ,["(lhs e_) is_sqrtequation_in v_ | \
		\(rhs e_) is_sqrtequation_in v_"]),
    ("#Find"  ,["solutions v_i_"])
   ],
   {rew_ord'="termlessI",
    rls'=rooteq_erls,
    srls=e_rls,
    prls=rooteq_prls,
    calc=[],
    asm_rls=["rooteq_simplify"],
    asm_thm=[]},
   "Script Norm_sq_root_equation  (e_::bool) (v_::real)  =      \
    \(let e_ = ((Repeat(Try (Rewrite     makex1_x           False))) @@  \
    \           (Try (Repeat (Rewrite_Set expand_rootbinoms False))) @@  \ 
    \           (Try (Rewrite_Set rooteq_simplify            False)) @@  \ 
    \           (Try (Repeat (Rewrite_Set make_rooteq       False))) @@  \
    \           (Try (Rewrite_Set rooteq_simplify            False))) e_ \
    \ in ((SubProblem (RootEq_,[sq,root,univariate,equation],      \
    \      (RootEq_,no_met)) [bool_ e_, real_ v_])))"
   )
,
prep_met
 (("RootEq.thy","solve_sq_root_equation"),
   [("#Given" ,["equality e_","solveFor v_"]),
    ("#Where" ,["(lhs e_) is_sqrtequation_in v_ | \
		\(rhs e_) is_sqrtequation_in v_"]),
    ("#Find"  ,["solutions v_i_"])
   ],
   {rew_ord'="termlessI",
    rls'=rooteq_erls,
    srls = rooteq_srls,
    prls = rooteq_prls,
    calc = [],
    asm_rls = ["sqrt_isolate","rooteq_simplify"],
    asm_thm = [("sqrt_square_1",""),("sqrt_square_equation_left_1",""),
               ("sqrt_square_equation_left_2",""),("sqrt_square_equation_left_3",""),
               ("sqrt_square_equation_left_4",""),("sqrt_square_equation_right_1",""),
               ("sqrt_square_equation_right_2",""),("sqrt_square_equation_right_3",""),
               ("sqrt_square_equation_right_4","")]},
    "Script Solve_sq_root_equation  (e_::bool) (v_::real)  =                      \
    \(let e_ = ((Try (Rewrite_Set_Inst [(bdv,v_::real)] sqrt_isolate    True)) @@ \
    \           (Try (Rewrite_Set                       rooteq_simplify True)) @@ \
    \           (Try (Repeat (Rewrite_Set expand_rootbinoms           False))) @@ \
    \           (Try (Repeat (Rewrite_Set make_rooteq                 False))) @@ \
    \           (Try (Rewrite_Set rooteq_simplify                       True))) e_;\
    \ (L_::bool list) =                                                            \
    \    (if (((lhs e_) is_sqrtequation_in v_) | ((rhs e_) is_sqrtequation_in v_)) \ 
    \ then (SubProblem (RootEq_,[normalize,root,univariate,equation],              \
    \       (RootEq_,no_met)) [bool_ e_, real_ v_])                                \
    \ else (SubProblem (RootEq_,[univariate,equation],                             \
    \        (RootEq_,no_met)) [bool_ e_, real_ v_]))                              \
    \ in Check_elementwise L_ {(v_::real). Assumptions})"
 )
,
(*-- right 28.08.02 --*)
prep_met
 (("RootEq.thy","solve_right_sq_root_equation"),
   [("#Given" ,["equality e_","solveFor v_"]),
    ("#Where" ,["(rhs e_) is_sqrtequation_in v_"]),
    ("#Find"  ,["solutions v_i_"])
   ],
   {rew_ord'="termlessI",
    rls'=rooteq_erls,
    srls=e_rls,
    prls=rooteq_prls,
    calc=[],
    asm_rls=["r_sqrt_isolate","rooteq_simplify"],
    asm_thm=[("sqrt_square_1",""),("sqrt_square_equation_right_1",""),
             ("sqrt_square_equation_right_2","")]},
  "Script Solve_right_sq_root_equation  (e_::bool) (v_::real)  =                   \
    \(let e_ = ((Try (Rewrite_Set_Inst [(bdv,v_::real)] r_sqrt_isolate  False)) @@ \       
    \           (Try (Rewrite_Set                       rooteq_simplify False)) @@ \ 
    \           (Try (Repeat (Rewrite_Set expand_rootbinoms            False))) @@ \
    \           (Try (Repeat (Rewrite_Set make_rooteq                  False))) @@ \
    \           (Try (Rewrite_Set rooteq_simplify                       False))) e_\
    \ in if ((rhs e_) is_sqrtequation_in v_)                                     \ 
    \ then (SubProblem (RootEq_,[normalize,root,univariate,equation],            \
    \       (RootEq_,no_met)) [bool_ e_, real_ v_])                              \
    \ else ((SubProblem (RootEq_,[univariate,equation],                          \
    \        (RootEq_,no_met)) [bool_ e_, real_ v_])))"
 )
,
(*-- left 28.08.02 --*)
prep_met
 (("RootEq.thy","solve_left_sq_root_equation"),
   [("#Given" ,["equality e_","solveFor v_"]),
    ("#Where" ,["(lhs e_) is_sqrtequation_in v_"]),
    ("#Find"  ,["solutions v_i_"])
   ],
   {rew_ord'="termlessI",
    rls'=rooteq_erls,
    srls=e_rls,
    prls=rooteq_prls,
    calc=[],
    asm_rls=["l_sqrt_isolate","rooteq_simplify"],
    asm_thm=[("sqrt_square_1",""),("sqrt_square_equation_left_1",""),
               ("sqrt_square_equation_left_2","")]},
    "Script Solve_left_sq_root_equation  (e_::bool) (v_::real)  =                  \
    \(let e_ = ((Try (Rewrite_Set_Inst [(bdv,v_::real)] l_sqrt_isolate  False)) @@ \
    \           (Try (Rewrite_Set                       rooteq_simplify False)) @@ \
    \           (Try (Repeat (Rewrite_Set expand_rootbinoms            False))) @@ \
    \           (Try (Repeat (Rewrite_Set make_rooteq                  False))) @@ \
    \           (Try (Rewrite_Set rooteq_simplify                       False))) e_\
    \ in if ((lhs e_) is_sqrtequation_in v_)                                     \ 
    \ then (SubProblem (RootEq_,[normalize,root,univariate,equation],            \
    \       (RootEq_,no_met)) [bool_ e_, real_ v_])                              \
    \ else ((SubProblem (RootEq_,[univariate,equation],                          \
    \        (RootEq_,no_met)) [bool_ e_, real_ v_])))"
   )
]);

"******* RootEq.ML end *******";
