(* (c) by Richard Lang
   collecting all knowledge for Root and Rational Equations
   created by: rlang 
         date: 02.10
   changed by: rlang
   last change by: rlang
             date: 02.11.04
*)

(* use"knowledge/RootRatEq.ML";
   use"RootRatEq.ML";

   use"ROOT.ML";
   cd"knowledge";

   remove_thy"RootRatEq";
   use_thy"Isac";
   *)

"******* RootRatEq.ML begin *******";
theory' := overwritel (!theory', [("RootRatEq.thy",RootRatEq.thy)]);

(*-------------------------functions---------------------*)
(* true if denominator contains (sq)root in + or - term 
   1/(sqrt(x+3)*(x+4)) -> false; 1/(sqrt(x)+2) -> true
   if false then (term)^2 contains no (sq)root *)
fun is_ratrootaddeq_in t v = 
    let 
	fun coeff_in c v = v mem (vars c);
	fun rootadd (t as (Const ("op +",_) $ t2 $ t3)) v = (is_rootequation_in t2 v) orelse 
	                                                    (is_rootequation_in t3 v)
	  | rootadd (t as (Const ("op -",_) $ t2 $ t3)) v = (is_rootequation_in t2 v) orelse 
                                                            (is_rootequation_in t3 v)
	  | rootadd _ _ = false;
	fun findratroot (_ $ _ $ _ $ _) v = raise error("is_ratrootaddeq_in:")
	  (* at the moment there is no term like this, but ....*)
	  | findratroot (t as (Const ("HOL.divide",_) $ _ $ t3)) v = 
	               if (is_rootequation_in t3 v) then rootadd t3 v else false
	  | findratroot _ _ = false;
     in
	findratroot t v
    end;

fun eval_is_ratrootaddeq_in _ _ (p as (Const ("RootRatEq.is'_ratrootaddeq'_in",_) $ t $ v)) _  =
    if is_ratrootaddeq_in t v then 
	Some ((term2str p) ^ " = True",
	      Trueprop $ (mk_equality (p, HOLogic.true_const)))
    else Some ((term2str p) ^ " = True",
	       Trueprop $ (mk_equality (p, HOLogic.false_const)))
  | eval_is_ratrootaddeq_in _ _ _ _ = ((*writeln"### nichts matcht";*) None);

(*-------------------------rulse-------------------------*)
val rootrateq_prls = 
    merge_rls "rootrateq_prls" rateq_prls 
    (merge_rls "rooteq_prls__" rooteq_prls
    (append_rls "is_ratrootaddeq_in" e_rls
		[Calc ("RootRatEq.is'_ratrootaddeq'_in",
		       eval_is_ratrootaddeq_in "")
		 ]));


val rootrateq_erls = 
    merge_rls "rootrateq_erls" rootrat_erls
    (merge_rls "" rooteq_erls
     (merge_rls "" rateq_erls
      (append_rls "" e_rls
		[])));

ruleset' := overwritel (!ruleset',
			[("rootrateq_erls",rootrateq_erls) (*FIXXXME:del with rls.rls'*) 
			 ]);

(* Solves a ratRoot Equation *)
 val ratroot_solve = 
     Rls {id = "ratroot_solve", preconds = [], 
	  rew_ord = ("termlessI",termlessI), 
     erls = e_rls, srls = Erls, calc = [], asm_thm = [],
     rules = [  Thm("rootrat_equation_left",num_str rootrat_equation_left),   
	        (* [|c is_rootequation_in bdv|] ==> ( (a + b/c = d) = ( b = (d - a) * c )) *)
	        Thm("rootrat_equation_right",num_str rootrat_equation_right)   
		(* [|f is_rootequation_in bdv|] ==> ( (a = d + e/f) = ( (a - d) * f = e )) *)
	      ],
	 scr = Script ((term_of o the o (parse thy)) "empty_script")
         }:rls;
ruleset' := overwritel (!ruleset',
			[("ratroot_solve",ratroot_solve)
			 ]);

(*-----------------------probleme------------------------*)
(*
(get_pbt ["rat","root","univariate","equation"]);
show_ptyps(); 
*)
store_pbt
 (prep_pbt RootRatEq.thy
 (["rat","root","univariate","equation"],
  [("#Given" ,["equality e_","solveFor v_"]),
   ("#Where" ,["( (lhs e_) is_ratrootaddeq_in (v_::real) )| \
	       \( (rhs e_) is_ratrootaddeq_in (v_::real) )"]),
   ("#Find"  ,["solutions v_i_"])
   ],
  rootrateq_prls, None,
  [("RootRatEq.thy","elim_rootrat_equation")]));

(*-------------------------Methode-----------------------*)
methods:= overwritel (!methods,
[
(*-- left 20.10.02 --*)
prep_met
 (("RootRatEq.thy","elim_rootrat_equation"),
   [("#Given" ,["equality e_","solveFor v_"]),
    ("#Where" ,["( (lhs e_) is_ratrootaddeq_in (v_::real) ) | \
	       \( (rhs e_) is_ratrootaddeq_in (v_::real) )"]),
    ("#Find"  ,["solutions v_i_"])
   ],
   {rew_ord'="termlessI",
    rls'=rootrateq_erls,
    srls=e_rls,
    prls=rootrateq_prls,
    calc=[],
    asm_rls=[],
    asm_thm=[]},
   "Script Elim_rootrat_equation  (e_::bool) (v_::real)  =      \
    \(let e_ = ((Try (Rewrite_Set expand_rootbinoms False)) @@  \ 
    \           (Try (Rewrite_Set rooteq_simplify   False)) @@  \ 
    \           (Try (Rewrite_Set make_rooteq       False)) @@  \
    \           (Try (Rewrite_Set rooteq_simplify   False)) @@  \
    \           (Try (Rewrite_Set_Inst [(bdv,v_)]               \
    \                                  ratroot_solve False))) e_ \
    \ in (SubProblem (RootEq_,[univariate,equation],            \
    \        (RootEq_,no_met)) [bool_ e_, real_ v_]))"
   )
]);
"******* RootRatEq.ML end *******";
