(* use"../ME/script.sml";
   use"ME/script.sml";
   use"script.sml";
   *)

signature INTERPRETER =
sig
  (*type ets (list of executed tactics) see sequent.sml*)

  datatype locate
    = NotLocatable
    | Steps of (mstep' * mout * ptree * pos' * cid * safe (* ets*)) list
(*    | ToDo of ets 28.4.02*)

  (*diss: next-tactic-function*)
  val next_tac : theory' -> ptree * pos' -> metID -> scr -> ets -> mstep
  (*diss: locate-function*)
  val locate_gen : theory'
                   -> mstep'
                      -> ptree * pos' -> scr * rls -> ets -> loc_ -> locate

  val sel_rules : ptree -> pos' -> mstep list
  val init_form : scr -> ets -> loc_ * term option
  val actual_args : term -> term list

  (*shift to library ...*)
  val inst_abs : theory' -> term -> term
  val itms2args : metID -> itm list -> term list
  val user_interrupt : loc_ * (mstep' * env * env * term * term * safe)
  (*val empty : term*) 
end 




(*
structure Interpreter : INTERPRETER =
struct
*)

val string_of_thm' = (implode o tl o drop_last o explode o string_of_thm);
(*
  string_of_thm sym;
val it = "\"?s = ?t ==> ?t = ?s\"" : string
  string_of_thm' sym;
val it = "?s = ?t ==> ?t = ?s" : string*)

type step =     (*data for creating a new node in the ptree;
		 designed for use:
               	 fun ass* scrstate steps =
               	 ... case ass* scrstate steps of
               	     Assoc (scrstate, steps) => ... ass* scrstate steps*)
    mstep'       (*transformed from associated tac*)
    * mout       (*result with indentation etc.*)
    * ptree      (*containing node created by mstep' + resp. scrstate*)
    * pos'       (*position in ptree; ptree * pos' is the proofstate*)
    * cid(*pos' list TODO*);(*of ptree-nodes ev. cut (by fst mstep')*)

fun rule2thm' (Thm (id, thm)) = (id, string_of_thm thm):thm'
  | rule2thm' r = raise error ("rule2thm': not defined for "^(rule2str r));

(*.makes a (rule,term) list to a Step (m, mout, pt', p', cid) for solve;
   complicated with current t in rrlsstate.*)
fun rts2steps steps ((pt,p),(f,f'',rss,rts),(thy',ro,er,pa)) [(r, (f', am))] =
    let val thy = assoc_thy thy'
	val m = Rewrite' (thy',ro,er,pa, rule2thm' r, f, (f', am))
	val is = RrlsState (f',f'',rss,rts)
	val p = case p of (p',Frm) => p | (p',Res) => (lev_on p',Res)
	val (p', cid, mout, pt') = generate1 thy m is p pt
    in (is, (m, mout, pt', p', cid)::steps) end
  | rts2steps steps ((pt,p),(f,f'',rss,rts),(thy',ro,er,pa)) 
	      ((r, (f', am))::rts') =
    let val thy = assoc_thy thy'
	val m = Rewrite' (thy',ro,er,pa, rule2thm' r, f, (f', am))
	val is = RrlsState (f',f'',rss,rts)
	val p = case p of (p',Frm) => p | (p',Res) => (lev_on p',Res)
	val (p', cid, mout, pt') = generate1 thy m is p pt
    in rts2steps ((m, mout, pt', p', cid)::steps) 
		 ((pt',p'),(f',f'',rss,rts),(thy',ro,er,pa)) rts' end;


val trace_locate = ref false; (*23.11.01*)

(*. functions for the environment stack .*)
fun accessenv id es = the (assoc((top es):env, id))
    handle _ => error ("accessenv: "^(free2str id)^" not in env");
fun updateenv id vl (es:env stack) = 
    (push (overwrite(top es, (id, vl))) (pop es)):env stack;
fun pushenv id vl (es:env stack) = 
    (push (overwrite(top es, (id, vl))) es):env stack;
val popenv = pop:env stack -> env stack;



(*
fun envETR2str thy' (_,(ens,env,iar,res,s)) = subst2str' thy' env;
fun resETR2str thy' (_,(ens,env,iar,res,s)) = 
  Sign.string_of_term (sign_of (assoc_thy thy')) res;----------------*)


(*val empty = Const ("empty",HOLogic.realT);
val user_interrupt = ([]:loc_,(User', []:env, []:env, empty, empty, Sundef));*)
fun de_esc_underscore str =
  let fun scan [] = []
	| scan (s::ss) = if s = "'" then (scan ss)
			 else (s::(scan ss))
  in (implode o scan o explode) str end;
(*
> val str = "Rewrite_Set_Inst";
> val esc = esc_underscore str;
val it = "Rewrite'_Set'_Inst" : string
> val des = de_esc_underscore esc;
 val des = de_esc_underscore esc;*)

fun is_listexpr t = 
  (((ids_of o head_of) t) inter (!listexpr)) <> [];


(*go at a location in a script and fetch the contents*)
fun go [] t = t
  | go (D::p) (Abs(s,ty,t0)) = go (p:loc_) t0
  | go (L::p) (t1 $ t2) = go p t1
  | go (R::p) (t1 $ t2) = go p t2
  | go l _ = raise error ("go: no "^(loc_2str l));
(*
> val t = (term_of o the o (parse thy)) "a+b";
val it = Const (#,#) $ Free (#,#) $ Free ("b","RealDef.real") : term
> val plus_a = go [L] t; 
> val b = go [R] t; 
> val plus = go [L,L] t; 
> val a = go [L,R] t;

> val t = (term_of o the o (parse thy)) "a+b+c";
val t = Const (#,#) $ (# $ # $ Free #) $ Free ("c","RealDef.real") : term
> val pl_pl_a_b = go [L] t; 
> val c = go [R] t; 
> val a = go [L,R,L,R] t; 
> val b = go [L,R,R] t; 
*)


(* get a subterm t with test t, and record location *)
fun get l test (t as Const (s,T)) = 
    if test t then Some (l,t) else None
  | get l test (t as Free (s,T)) = 
    if test t then Some (l,t) else None 
  | get l test (t as Bound n) =
    if test t then Some (l,t) else None 
  | get l test (t as Var (s,T)) =
    if test t then Some (l,t) else None
  | get l test (t as Abs (s,T,body)) =
    if test t then Some (l:loc_,t) else get ((l@[D]):loc_) test body
  | get l test (t as t1 $ t2) =
    if test t then Some (l,t) 
    else case get (l@[L]) test t1 of 
      None => get (l@[R]) test t2
    | Some (l',t') => Some (l',t');
(*18.6.00
> val sss = ((term_of o the o (parse thy))
  "Script Solve_root_equation (eq_::bool) (v_::real) (err_::bool) =\
   \ (let e_ = Try (Rewrite square_equation_left True eq_) \
   \  in [e_])");
          ______ compares head_of !!
> get [] (eq_str "Let") sss;            [R]
> get [] (eq_str "Script.Try") sss;     [R,L,R]
> get [] (eq_str "Script.Rewrite") sss; [R,L,R,R]
> get [] (eq_str "True") sss;           [R,L,R,R,L,R]
> get [] (eq_str "e_") sss;             [R,R]
*)

fun test_negotiable t = ((strip_thy o (term_str Script.thy) o head_of) t) 
  mem (!negotiable);

(*30.4.02: vvv--- doesnt work with curried functions ---> get_tac ------
(*18.6.00: below _ALL_ negotiables must be in fun-patterns !
  then the last (non)pattern must be a subproblem*)
fun init_frm thy (Const ("Script.Rewrite",_) $ _ $ _ $ eq) = Some eq
  | init_frm thy (Const ("Script.Rewrite'_Inst",_) $ _ $ _ $ _ $ eq) = Some eq
  | init_frm thy (Const ("Script.Rewrite'_Set",_) $ _ $ _ $ eq) = Some eq
  | init_frm thy (Const ("Script.Rewrite'_Set'_Inst",_) $ _ $ _ $ _ $ eq) = 
    Some eq
  | init_frm thy (Const ("Script.Calculate",_) $ _ $ t) = Some t
  | init_frm thy t = 
  (*if ((strip_thy o (term_str thy) o head_of) t) mem (!subpbls)
    then None 
  else *)raise error ("init_frm: not impl. for "^
		    (Sign.string_of_term (sign_of thy) t));

> val t = (term_of o the o (parse thy)) 
 "Rewrite square_equation_left True (sqrt(#9+#4*x)=sqrt x + sqrt(#5+x))";
> val Some ini = init_frm thy t;
> Sign.string_of_term (sign_of thy) ini;
val it = "sqrt (#9 + #4 * x) = sqrt x + sqrt (#5 + x)" : string

> val t = (term_of o the o (parse thy)) 
 "solve_univar (Reals, [univar,equation], no_met) e1_ v1_";
> val ini = init_frm thy t;
> Sign.string_of_term (sign_of thy) ini;
val it = "empty" : string

> val t = (term_of o the o (parse thy)) 
 "Rewrite_Set norm_equation False x + #1 = #2";
> val Some ini = init_frm thy t;
> Sign.string_of_term (sign_of thy) ini;
val it = "x + #1 = #2" : string                                                

> val t = (term_of o the o (parse thy)) 
 "Rewrite_Set_Inst [(bdv,x)] isolate_bdv False x + #1 = #2";
> val Some ini = init_frm thy t;
> Sign.string_of_term (sign_of thy) ini;
val it = "x + #1 = #2" : string                           *)


(*get argument of first tactic in a script for init_form*)
fun get_tac thy (h $ body) =
  let
    fun get_t y (Const ("Script.Seq",_) $ e1 $ e2) a = 
    	(case get_t y e1 a of None => get_t y e2 a | la => la)
      | get_t y (Const ("Script.Seq",_) $ e1 $ e2 $ a) _ = 
    	(case get_t y e1 a of None => get_t y e2 a | la => la)
      | get_t y (Const ("Script.Try",_) $ e) a = get_t y e a
      | get_t y (Const ("Script.Try",_) $ e $ a) _ = get_t y e a
      | get_t y (Const ("Script.Repeat",_) $ e) a = get_t y e a
      | get_t y (Const ("Script.Repeat",_) $ e $ a) _ = get_t y e a
      | get_t y (Const ("Script.Or",_) $e1 $ e2) a =
    	(case get_t y e1 a of None => get_t y e2 a | la => la)
      | get_t y (Const ("Script.Or",_) $e1 $ e2 $ a) _ =
    	(case get_t y e1 a of None => get_t y e2 a | la => la)
      | get_t y (Const ("Script.While",_) $ c $ e) a = get_t y e a
      | get_t y (Const ("Script.While",_) $ c $ e $ a) _ = get_t y e a
      | get_t y (Const ("Script.Letpar",_) $ e1 $ Abs (_,_,e2)) a = 
    	(case get_t y e1 a of None => get_t y e2 a | la => la)
      | get_t y (Const ("Let",_) $ e1 $ Abs (_,_,e2)) a =
    	(case get_t y e1 a of None => get_t y e2 a | la => la)
      | get_t y (Const ("If",_) $ c $ e1 $ e2) a = 
    	(case get_t y e1 a of None => get_t y e2 a | la => la)
      | get_t y (Abs (_,_,e)) a = get_t y e a
    
      | get_t y (Const ("Script.Rewrite",_) $ _ $ _ $ a) _ = Some a
      | get_t y (Const ("Script.Rewrite",_) $ _ $ _    ) a = Some a
      | get_t y (Const ("Script.Rewrite'_Inst",_) $ _ $ _ $ _ $ a) _ = Some a
      | get_t y (Const ("Script.Rewrite'_Inst",_) $ _ $ _ $ _ )    a = Some a
      | get_t y (Const ("Script.Rewrite'_Set",_) $ _ $ _ $ a) _ = Some a
      | get_t y (Const ("Script.Rewrite'_Set",_) $ _ $ _ )    a = Some a
      | get_t y (Const ("Script.Rewrite'_Set'_Inst",_) $ _ $ _ $ _ $a)_ =Some a
      | get_t y (Const ("Script.Rewrite'_Set'_Inst",_) $ _ $ _ $ _ )  a =Some a
      | get_t y (Const ("Script.Calculate",_) $ _ $ a) _ = Some a
      | get_t y (Const ("Script.Calculate",_) $ _ )    a = Some a
    
      | get_t y (Const ("Script.Substitute",_) $ _ $ a) _ = Some a
      | get_t y (Const ("Script.Substitute",_) $ _ )    a = Some a
    
      | get_t y (Const ("Script.SubProblem",_) $ _ $ _) _ = None

      | get_t y x _ =  raise error ("get_t yac: not defined for "^
    				  (Sign.string_of_term (sign_of y) x))
in get_t thy body e_term end;
    
(*FIXME: get 1st tac by next_tac [] instead of ... ?? 29.7.02*)
(* val (Script sc,((_,(_,_,env,_,_,_))::_)) = (scr,ets);
   *)
fun init_form thy (Script sc) env =
  (case get_tac thy sc of
     None => None (*raise error ("init_form: no 1st tac in "^
			  (Sign.string_of_term (sign_of thy) sc))*)
   | Some tac => Some (subst_atomic env tac))
  | init_form _ _ _ = raise error "init_form: no match";

(* use"ME/script.sml";
   use"script.sml";
   *)



(*the 'iteration-argument' of a tac (args not eval)*)
fun itr_arg _ (Const ("Script.Rewrite'_Inst",_) $ _ $ _ $ _ $ v) = v
  | itr_arg _ (Const ("Script.Rewrite",_) $ _ $ _ $ v) = v
  | itr_arg _ (Const ("Script.Rewrite'_Set'_Inst",_) $ _ $ _ $ _ $ v) = v
  | itr_arg _ (Const ("Script.Rewrite'_Set",_) $ _ $ _ $ v) = v
  | itr_arg _ (Const ("Script.Calculate",_) $ _ $ v) = v
  | itr_arg _ (Const ("Script.Check'_elementwise",_) $ consts $ _) = consts
  | itr_arg _ (Const ("Script.Or'_to'_List",_) $ _) = e_term
  | itr_arg _ (Const ("Script.Mstep",_) $ _) = e_term
  | itr_arg _ (Const ("Script.SubProblem",_) $ _ $ _) = e_term
  | itr_arg thy t = raise error 
    ("itr_arg not impl. for "^
     (Sign.string_of_term (sign_of (assoc_thy thy)) t));
(* val t = (term_of o the o (parse thy))"Rewrite rroot_square_inv False e_";
> itr_arg "Script.thy" t;
val it = Free ("e_","RealDef.real") : term 
> val t = (term_of o the o (parse thy))"xxx";
> itr_arg "Script.thy" t;
*** itr_arg not impl. for xxx
uncaught exception ERROR
  raised at: library.ML:1114.35-1114.40*)


fun actual_args scr = (fst o split_last o snd o strip_comb) scr;
(*
> parms scr;
  [Free ("f_","RealDef.real"),Free ("v_","RealDef.real"),
   Free ("eqs_","bool List.list")] : term list
*)


(*update environment; t <> empty if coming from listexpr*)
fun upd_env (env:env) (v,t) =
  let val env' = if t = e_term then env else overwrite (env,(v,t));
    (*val _= writeln("### upd_env: = "^(subst2str env'));*)
  in env' end;

(*26.5.02: not clear, when a is available in ass_up for eva-_true*)
fun upd_env_opt env (Some a, v) = upd_env env (a,v)
  | upd_env_opt env (None, v) = 
    (writeln("*** upd_env_opt: (None,"^(term2str v)^")");env);


type dsc = typ; (*<-> nam..unknow in Descript.thy*)
fun typ_str (Type (s,_)) = s
  | typ_str (TFree(s,_)) = s
  | typ_str (TVar ((s,i),_)) = s^(string_of_int i);
	     
(*get the _result_-type of a description*)
fun dsc_valT (Const (_,(Type (_,[_,T])))) = (strip_thy o typ_str) T;
(*> val t = (term_of o the o (parse thy)) "equality";
> val T = type_of t;
val T = "bool => Tools.una" : typ
> val dsc = dsc_valT t;
val dsc = "una" : string

> val t = (term_of o the o (parse thy)) "fixedValues";
> val T = type_of t;
val T = "bool List.list => Tools.nam" : typ
> val dsc = dsc_valT t;
val dsc = "nam" : string*)

(*make args for script depending on type of description*)
fun mk_arg thy d [] = raise error ("mk_arg: no data for "^
			       (Sign.string_of_term (sign_of thy) d))
  | mk_arg thy d [t] = 
  (case dsc_valT d of
    "una" => [t]
  | "nam" => 
      [case t of
	 r as (Const ("op =",_) $ _ $ _) => r
       | _ => raise error 
	   ("mk_arg: dsc-typ 'nam' applied to non-equality "^
	    (Sign.string_of_term (sign_of thy) t))]
  | s => raise error ("mk_arg: not impl. for "^s))

  | mk_arg thy d (t::ts) = (mk_arg thy d [t]) @ (mk_arg thy d ts);
(* 
 val d = d_in itm_;
 val [t] = ts_in itm_;
mk_arg thy

*)


(*create the actual parameters (args) of script: their order 
  is given by the order of "#Given" in ppc in !methods*)
(* val mI = methID;
   *)
fun itms2args thy mI (itms:itm list) =
  let
    fun test_dsc d (_,_,_,_,itm_) = (d = d_in itm_);
    fun itm2arg itms (_,(d,_)) =
      case find_first (test_dsc d) itms of
	None => 
	  raise error ("itms2args: '"^
		       (Sign.string_of_term (sign_of thy) d)^
		       "' not in itms")
      | Some (_,_,_,_,itm_) => mk_arg thy (d_in itm_) (ts_in itm_);
    fun test_given (s,_) = (s = "#Given");
    val mpc = (#ppc o get_met) mI;
    val gs = filter test_given mpc;
  in (flat o (map (itm2arg itms))) gs end;
(*
> val sc = ... Solve_root_equation ...
> val mI = ("Script.thy","sqrt-equ-test");
> val PblObj{meth={ppc=itms,...},...} = get_obj I pt [];
> val ts = itms2args thy mI itms;
> map (Sign.string_of_term (sign_of thy)) ts;
["sqrt (#9 + #4 * x) = sqrt x + sqrt (#5 + x)","x","#0"] : string list
*)



(*detour necessary, because generate1 delivers a string-result*)
fun mout2term thy (Form' (FormKF (_,_,_,_,res))) = 
  (term_of o the o (parse (assoc_thy thy))) res
  | mout2term thy (Form' (PpcKF _)) = e_term;(*3.8.01: res of subpbl 
					   at time of detection in script*)

(* gets data from script which are necessary for next_tac (=mstep);
   (other data are fetched from pt in 'appl_in')
   12.1.01: returnvalue term is useful for appl_in 
   26.9.00: still necessary? ~~was ? *)
fun tac2mstep' thy (Const ("Script.Rewrite",_) $ Free (thmID,_) $ _ $ f) =
    let val tid = (de_esc_underscore o strip_thy) thmID
    in (Rewrite (tid, (de_quote o string_of_thm o 
		       (assoc_thm' thy)) (tid,"")), f) end
(* val (thy,
	mm as(Const ("Script.Rewrite'_Inst",_) $  sub $ Free(thmID,_) $ _ $ f))
     = (assoc_thy th,tac);
   tac2mstep' thy mm;

   assoc_thm' (assoc_thy "Isac.thy") (tid,"");
   assoc_thm' Isac.thy (tid,"");
   *)
  | tac2mstep' thy (Const ("Script.Rewrite'_Inst",_) $ 
	       sub $ Free (thmID,_) $ _ $ f) =
  let val subML = ((map isapair2pair) o isalist2list) sub
    val subStr = subst2subs subML
    val tid = (de_esc_underscore o strip_thy) thmID (*4.10.02 unnoetig*)
  in (Rewrite_Inst 
	  (subStr, (tid, (de_quote o string_of_thm o
			  (assoc_thm' thy)) (tid,""))), f) end
      
  | tac2mstep' thy (Const ("Script.Rewrite'_Set",_) $ Free (rls,_) $ _ $ f)=
  (Rewrite_Set ((de_esc_underscore o strip_thy) rls), f)

  | tac2mstep' thy (Const ("Script.Rewrite'_Set'_Inst",_) $ 
	       sub $ Free (rls,_) $ _ $ f) =
  let val subML = ((map isapair2pair) o isalist2list) sub;
    val subStr = subst2subs subML;
  in (Rewrite_Set_Inst (subStr,rls), f) end

  | tac2mstep' thy (Const ("Script.Calculate",_) $ Free (op_,_) $ f) =
  (Calculate op_, f)

(*12.1.01.*)
  | tac2mstep' thy (Const("Script.Check'_elementwise",_) $ _ $ 
		    (set as Const ("Collect",_) $ Abs (_,_,pred))) = 
  (Check_elementwise (Sign.string_of_term (sign_of thy) pred), set)

  | tac2mstep' thy (Const("Script.Or'_to'_List",_) $ _ ) = 
  (Or_to_List, e_term)

(*12.1.01.for subproblem_equation_dummy in root-equation *)
  | tac2mstep' thy (Const ("Script.Mstep",_) $ Free (str,_)) = 
  (Mstep ((de_esc_underscore o strip_thy) str), e_term) 
		    (*L_ will come from pt in appl_in*)

  | tac2mstep' thy (Const ("Script.SubProblem",_) $
	(Const ("Pair",_) $ Free (thy',_) $
	       (Const ("Pair",_) $ pblID' $
		      (Const ("Pair",_) $ Free (mI1',_) $ Free (mI2',_)))) $
	ags') =
    let val pblID = ((map (de_esc_underscore o free2str)) 
		     o isalist2list) pblID'
	val thy_ = ((implode o drop_last o explode) thy')^".thy"
    in (Subproblem (thy_, pblID), subpbl thy' pblID) end

  | tac2mstep' thy t = raise error 
  ("tac2mstep' TODO: no match for "^
   (Sign.string_of_term (sign_of thy) t));
(*
> val t = (term_of o the o (parse thy)) 
 "Rewrite_Set_Inst [(bdv,v_::real)] isolate_bdv False (x=a+#1)";
> tac2mstep' t;
val it = Rewrite_Set_Inst ([(#,#)],"isolate_bdv") : mstep

> val t = (term_of o the o (parse SqRoot.thy)) 
"(SubProblem (SqRoot_,[equation,univariate],(SqRoot_,solve_linear))\
   \         [bool_ e_, real_ v_])::bool list";
> tac2mstep' SqRoot.thy t;
val it = (Subproblem ("SqRoot.thy",[#,#]),Const (#,#) $ (# $ # $ (# $ #)))
*)

fun tac2mstep thy t = (fst o tac2mstep' thy) t;

(*24.6.02 unused, use subst_tac instead !*)
fun is_tac t =
    (tac2mstep' Script.thy t; true)
    handle _ => false;


(* instantiate a tactic, and ev. attach (curried) argument
args:
   E       environment
   v       current value, is attached to curried tactics
   tac     tactic to be instantiated
precond:
   not (a = None) /\ (v = e_term) /\ (tac curried, i.e. without last arg.)
   this ........................ is the initialization for assy with l=[],
   but the 1st tac is
   (a) curried:     then (a = Some _), or 
   (b) not curried: then the values of the initialization are not used
*)
fun subst_tac E a v (t as (Const ("Script.Rewrite",_) $ _ $ _ $ _ ))=
    subst_atomic (case a of Some a => upd_env E (a,v) | None => E) t

  | subst_tac E a v (t as (Const ("Script.Rewrite",_) $ _ $ _ ))=
    (case a of Some a' => (subst_atomic E (t $ a'))
	     | None => ((subst_atomic E t) $ v))

  | subst_tac E a v 
	      (t as (Const ("Script.Rewrite'_Inst",_) $ _ $ _ $ _ $ _ )) =
    subst_atomic (case a of Some a => upd_env E (a,v) | None => E) t

  | subst_tac E a v (t as (Const ("Script.Rewrite'_Inst",_) $ _ $ _ $ _ )) =
    (case a of Some a' => subst_atomic E (t $ a')
	     | None => ((subst_atomic E t) $ v))

  | subst_tac E a v (t as (Const ("Script.Rewrite'_Set",_) $ _ $ _ $ _ )) =
    subst_atomic (case a of Some a => upd_env E (a,v) | None => E) t

  | subst_tac E a v (t as (Const ("Script.Rewrite'_Set",_) $ _ $ _ )) =
    (case a of Some a' => subst_atomic E (t $ a')
	     | None => ((subst_atomic E t) $ v))

  | subst_tac E a v 
	      (t as (Const ("Script.Rewrite'_Set'_Inst",_) $ _ $ _ $ _ $ _ )) =
    subst_atomic (case a of Some a => upd_env E (a,v) | None => E) t

  | subst_tac E a v 
	      (t as (Const ("Script.Rewrite'_Set'_Inst",_) $ _ $ _ $ _ )) =
    (case a of Some a' => subst_atomic E (t $ a')
	     | None => ((subst_atomic E t) $ v))

  | subst_tac E a v (t as (Const ("Script.Calculate",_) $ _ $ _ )) =
    subst_atomic (case a of Some a => upd_env E (a,v) | None => E) t

  | subst_tac E a v (t as (Const ("Script.Calculate",_) $ _ )) =
    (case a of Some a' => subst_atomic E (t $ a')
	     | None => ((subst_atomic E t) $ v))

  | subst_tac E a v 
	      (t as (Const("Script.Check'_elementwise",_) $ _ $ _ )) = 
    subst_atomic (case a of Some a => upd_env E (a,v) | None => E) t
(* val t = (term_of o the o (parseold Isac.thy))
		 "Check_elementwise L_ {v_. Assumptions}";
   val tactac = subst_atomic E t;
   atomt tactac;
   *)

  | subst_tac E a v (t as (Const("Script.Check'_elementwise",_) $ _ )) = 
    (case a of Some a' => subst_atomic E (t $ a')
	     | None => ((subst_atomic E t) $ v))

  | subst_tac E a v (t as (Const("Script.Or'_to'_List",_) $ _ )) = t

  | subst_tac E a v (t as (Const("Script.Or'_to'_List",_))) = (*t $ v*)
    (case a of Some a' => subst_atomic E (t $ a')
	     | None => ((subst_atomic E t) $ v))

  | subst_tac E a v (t as (Const ("Script.SubProblem",_) $
	(Const ("Pair",_) $ Free (thy',_) $
	       (Const ("Pair",_) $ pblID' $
		      (Const ("Pair",_) $ Free (mI1',_) $ Free (mI2',_)))) $
	ags')) =
    subst_atomic (case a of Some a => upd_env E (a,v) | None => E) t

  | subst_tac _ _ _ t = raise error 
  ("subst_tac TODO: no match for "^(term2str t));





(*start_at: go to the end of executed tactics; this is
  - the last element LE in ets, except this is in a let:
  - the 1st unfinished subexpr. in a let containing LE -
    within this subexpr. the last elem. wrt. iterations
  and return the location, local env, and 
  ets (containing ev. parallel tacs/subexpr. in let)

  language restriction: _sequential_ let's ONLY !
  miniversion below expects l= l of (last_elem ets)

args:
  l  : loc_ which generated given formula: for || let
  ets: from PblObj
  sc : script not yet(?) used
returns:
  enr: env, res etc. to be passed over until completion and append to ets
       27.4.01: adapt to diss:script-interpreter !!!
  ets: in future versions eventually modified (|| let) TODO

fun start_at ([]:loc_) ([]:ets) (sc:term) =
  (([]:env, []:env, empty, empty, Sundef), []:ets)
  (*neither l nor ets should be =[] ...*)
  | start_at l ets sc =
    let val (_,(_, ens, env, iar, ret, saf)) = last_elem (drop_last ets)
    (*TODO 5.9.00: check several 'User' instead ^^^^^^^^^*) 
    in ((ens, env, iar, ret, saf), ets) end;-----------------------3.5.02*)




fun tacpbls (h $ body) =
  let
    fun scan ts (Const ("Let",_) $ e $ (Abs (v,T,b))) =
      (scan ts e) @ (scan ts b)
      | scan ts (Const ("If",_) $ c $ e1 $ e2) = (scan ts e1) @ (scan ts e2)
      | scan ts (Const ("Script.While",_) $ c $ e $ _) = scan ts e
      | scan ts (Const ("Script.While",_) $ c $ e) = scan ts e
      | scan ts (Const ("Script.Repeat",_) $ e $ _) = scan ts e
      | scan ts (Const ("Script.Repeat",_) $ e) = scan ts e
      | scan ts (Const ("Script.Try",_) $ e $ _) = scan ts e
      | scan ts (Const ("Script.Try",_) $ e) = scan ts e
      | scan ts (Const ("Script.Or",_) $e1 $ e2 $ _) = 
	(scan ts e1) @ (scan ts e2)
      | scan ts (Const ("Script.Or",_) $e1 $ e2) = 
	(scan ts e1) @ (scan ts e2)
      | scan ts (Const ("Script.Seq",_) $e1 $ e2 $ _) = 
	(scan ts e1) @ (scan ts e2)
      | scan ts (Const ("Script.Seq",_) $e1 $ e2) = 
	(scan ts e1) @ (scan ts e2)
      | scan ts t = if is_listexpr t then [] else [t];
  in (distinct o (scan [])) body end;
    (*sc = Solve_root_equation ...
> val ts = tacpbls sc;
> writeln (terms2str thy ts);
["Rewrite square_equation_left True e_",
 "Rewrite_Set SqRoot_simplify False e_",
 "Rewrite_Set rearrange_assoc False e_",
 "Rewrite_Set isolate_root False e_",
 "Rewrite_Set norm_equation False e_",
 "Rewrite_Set_Inst [(bdv, v_)] isolate_bdv False e_"]
*)


(*get the tactics and problems of a script as msteps
  instantiated with the current environment;
  l is the location which generated the given formula *)
fun is_spec_pos Pbl = true
  | is_spec_pos Met = true
  | is_spec_pos _ = false;

fun sel_rules pt ((p,p_):pos') =
  if is_spec_pos p_ then [Mstep "sel_rules not impl. for specify"](*TODO*)
  else
    let val pp = par_pblobj pt p;
	val thy' = (get_obj g_domID pt pp):theory';
	val thy = assoc_thy thy';
	val metID = get_obj g_metID pt pp;
	val metID' = if metID =e_metID then(thd3 o snd)(get_obj g_origin pt pp)
		     else metID;
        val Script sc = (#scr o get_met) metID';
	val ScrState (env,_,_,_,_,_) = get_istate pt (p,p_);
    in map ((tac2mstep thy) o (subst_tac env None e_term)) (tacpbls sc) end;
(*
> val Script sc = (#scr o get_met) ("SqRoot.thy","sqrt-equ-test");
> val env = [((term_of o the o (parse Isac.thy)) "bdv",
             (term_of o the o (parse Isac.thy)) "x")];
> map ((tac2mstep thy) o (subst_tac env None e_term)) (tacpbls sc);
*)





(*test a term for being a _list_ (set ?) of constants; could be more rigorous*)
fun list_of_consts (Const ("List.list.Cons",_) $ _ $ _) = true
  | list_of_consts (Const ("List.list.Nil",_)) = true
  | list_of_consts _ = false;
(*val ttt = (term_of o the o (parse thy)) "[x=#1,x=#2,x=#3]";
> list_of_consts ttt;
val it = true : bool
> val ttt = (term_of o the o (parse thy)) "[]";
> list_of_consts ttt;
val it = true : bool*)





(* 15.1.01: evaluation of preds only works occasionally,
            but luckily for the 2 examples of root-equ:
> val s = ((term_of o the o (parse thy)) "x",
	   (term_of o the o (parse thy)) "-#5//#12");
> val asm = (term_of o the o (parse thy)) 
             "#0 <= #9 + #4 * x  &  #0 <= sqrt x + sqrt (#-3 + x)";
> val pred = subst_atomic [s] asm;
> rewrite_set_ thy false (cterm_of (sign_of thy) pred);
val it = None : (cterm * cterm list) option !!!!!!!!!!!!!!!!!!!!!!!!!!!!
> eval_true' (string_of_thy thy) "eval_rls" (subst_atomic [s] pred);
val it = false : bool

> val s = ((term_of o the o (parse thy)) "x",
	   (term_of o the o (parse thy)) "#4");
> val asm = (term_of o the o (parse thy)) 
             "#0 <= #9 + #4 * x  &  #0 <= sqrt x + sqrt (#5 + x)";
> val pred = subst_atomic [s] asm;
> rewrite_set_ thy false (cterm_of (sign_of thy) pred);
val it = Some ("True & True",[]) : (cterm * cterm list) option
> eval_true' (string_of_thy thy) "eval_rls" (subst_atomic [s] pred);
val it = true : bool`*)

(*for check_elementwise: take apart the set, ev. instantiate assumptions*)
fun rep_set thy pt p (set as Const ("Collect",_) $ Abs _) =
  let val (_ $ Abs (bdv,T,pred)) = inst_abs thy set;
    val bdv = Free (bdv,T);
    val pred = if pred <> Const ("Script.Assumptions",bool)
		 then pred 
	       else (mk_and o (map (term_of o the o (parse thy))) o 
		     (map fst)) (get_obj g_asm pt (par_pblobj pt p)) 
  in (bdv, pred) end
  | rep_set thy _ _ set = 
    raise error ("check_elementwise: no set "^ (*from script*)
		 (Sign.string_of_term (sign_of thy) set));
(*> val set = (term_of o the o (parse thy)) "{(x::real). Assumptions}";
> val p = [];
> val pt = union_asm pt p [("#0 <= sqrt x + sqrt (#5 + x)",[11]),
                           ("#0 <= #9 + #4 * x",[22]),
			   ("#0 <= x ^^^ #2 + #5 * x",[33]),
			   ("#0 <= #2 + x",[44])];
> val (bdv,pred) = rep_set thy pt p set;
val bdv = Free ("x","RealDef.real") : term
> writeln (Sign.string_of_term (sign_of thy) pred);
((#0 <= sqrt x + sqrt (#5 + x) & #0 <= #9 + #4 * x) &
 #0 <= x ^^^ #2 + #5 * x) &
#0 <= #2 + x
*)




datatype ass = 
  Ass of mstep' *  (*SubProblem gets args instantiated in assod*)
	 term      (*for itr_arg,result in ets*)
| AssWeak of mstep' *
	     term  (*for itr_arg,result in ets*)
| NotAss;

(*assod: mstep' associated with tac w.r.t. d
returns
 Ass    : associated: e.g. thmID in tac = thmID in m
                       +++ arg   in tac = arg   in m
 AssWeak: weakly ass.:e.g. thmID in tac = thmID in m, //arg//
 NotAss :             e.g. thmID in tac/=/thmID in m (not =)
8.01:
 mstep' SubProblem with args completed from script
*)
fun assod d (m as Rewrite_Inst' (thy',rod,rls,put,subs,(thmID,thm),f,(f',asm)))
  (Const ("Script.Rewrite'_Inst",_) $ subs_ $ Free (thmID_,idT) $ b $ f_) = 
   if thmID = thmID_ then 
    if f = f_ then ((*writeln"3### assod ..Ass";*)Ass (m,f')) 
    else ((*writeln"3### assod ..AssWeak";*)AssWeak(m, f'))
  else ((*writeln"3### assod ..NotAss";*)NotAss)

  | assod d (m as Rewrite' (thy,rod,rls,put,(thmID,thm),f,(f',asm))) 
  (t as Const ("Script.Rewrite",_) $ Free (thmID_,idT) $ b $ f_) =
  ((*writeln("3### assod: tac = "^
	   (Sign.string_of_term (sign_of (assoc_thy thy)) t));
   writeln("3### assod: f(m)= "^
	   (Sign.string_of_term (sign_of (assoc_thy thy)) f));*)
   if thmID = thmID_ then 
    if f = f_ then ((*writeln"3### assod ..Ass";*)Ass (m,f')) 
    else ((*writeln"### assod ..AssWeak";
	  writeln("### assod: f(m)  = "^
		  (Sign.string_of_term (sign_of (assoc_thy thy)) f));
	  writeln("### assod: f(tac)= "^
		  (Sign.string_of_term (sign_of (assoc_thy thy)) f_));*)
	  AssWeak (m,f'))
  else ((*writeln"3### assod ..NotAss";*)NotAss))

(*val f = (term_of o the o (parse thy))"#0+(sqrt(sqrt(sqrt a))^^^#2)^^^#2=#0";
> val f'= (term_of o the o (parse thy))"#0+(sqrt(sqrt a))^^^#2=#0";
> val m =   Rewrite'("Script.thy","tless_true","eval_rls",false,
 ("rroot_square_inv",""),f,(f',[]));
> val tac = (term_of o the o (parse thy))
 "Rewrite rroot_square_inv False (#0+(sqrt(sqrt(sqrt a))^^^#2)^^^#2=#0)";
> assod e_rls m tac;
val it =
  (Some (Rewrite' (#,#,#,#,#,#,#)),Const ("empty","RealDef.real"),
   Const ("empty","RealDef.real")) : mstep' option * term * term*)

  | assod d (m as Rewrite_Set_Inst' (thy',put,sub,rls,f,(f',asm))) 
  (Const ("Script.Rewrite'_Set'_Inst",_) $ sub_ $ Free (rls_,_) $ _ $ f_)= 
  if id_rls rls = rls_ then 
    if f = f_ then Ass (m,f') else AssWeak (m,f')
  else NotAss

  | assod d (m as Rewrite_Set' (thy,put,rls,f,(f',asm))) 
  (Const ("Script.Rewrite'_Set",_) $ Free (rls_,_) $ _ $ f_) = 
  if id_rls rls = rls_ then 
    if f = f_ then Ass (m,f') else AssWeak (m,f')
  else NotAss

  | assod d (m as Calculate' (thy',op_,f,(f',thm'))) 
  (Const ("Script.Calculate",_) $ Free (op__,_) $ f_) = 
  if op_ = op__ then
    if f = f_ then Ass (m,f') else AssWeak (m,f')
  else NotAss

  | assod _ (m as Check_elementwise' (consts,_,consts_chkd))
    (Const ("Script.Check'_elementwise",_) $ consts' $ _) =
    ((*writeln("### assod Check'_elementwise: consts = "^
	     (Sign.string_of_term (sign_of thy) consts));*)
     if consts = consts' then Ass (m,consts_chkd)
     else NotAss)

  | assod _ (m as Or_to_List' (ors, list)) 
	  (Const ("Script.Or'_to'_List",_) $ _) =
	  Ass (m, list) 

  | assod _ (m as Mstep' (thy,f,id,f'))  
    (Const ("Script.Mstep",_) $ Free (id',_)) =
    if id = id' then Ass (m, ((term_of o the o (parse thy)) f'))
    else NotAss

  | assod _ (Subproblem' ((domID,pblID,metID),_,f))
	  (Const ("Script.SubProblem",_) $
	(Const ("Pair",_) $ Free (dI',_) $
	       (Const ("Pair",_) $ pI' $
		      (Const ("Pair",_) $ Free (mI1',_) $ Free (mI2',_)))) $
	ags') =
    let val domID' = ((implode o drop_last o explode) dI')^".thy";
	val pblID' = ((map (de_esc_underscore o free2str)) 
		     o isalist2list) pI';
        val metID' as (dI',_) = (((implode o drop_last o explode) mI1')^".thy",
				  de_esc_underscore mI2');
	    (*5.8.02: really take domID for metID ?*)
	val oris = match_ags (assoc_thy domID) ((#ppc o get_pbt) pblID) 
			     (isalist2list ags');
    in
(*writeln("##### assoc: metID' = "^(metID2str metID'));*)
      if domID = domID' andalso pblID = pblID'
      then Ass (Subproblem' ((domID, pblID, metID'), oris, f), f) 
      else NotAss
    end

  | assod d m t = 
    (
(*writeln("### assod: NotAss m= "^(mstep'2str m)^"\n tac= "^(term2str t));*)
     NotAss);



fun mstep'2mstep (Rewrite' (thy,rod,erls,put,(thmID,thm),f,(f',asm))) =
  Rewrite (thmID,thm)

  | mstep'2mstep (Rewrite_Inst' (thy,rod,erls,put,sub,(thmID,thm),f,(f',asm)))=
  Rewrite_Inst (subst2subs sub,(thmID,thm))

  | mstep'2mstep (Rewrite_Set' (thy,put,rls,f,(f',asm))) = 
  Rewrite_Set (id_rls rls)

  | mstep'2mstep (Rewrite_Set_Inst' (thy,put,sub,rls,f,(f',asm))) = 
  Rewrite_Set_Inst (subst2subs sub,id_rls rls)

  | mstep'2mstep (Calculate' (thy,op_,t,(t',thm'))) = Calculate (op_)

  | mstep'2mstep (Check_elementwise' (consts,pred,consts')) =
  Check_elementwise pred

  | mstep'2mstep (Or_to_List' _) = Or_to_List

  | mstep'2mstep (Mstep' (_,f,id,f')) = Mstep id

  | mstep'2mstep (Subproblem' ((domID, pblID, _), _, _)) = 
		  Subproblem (domID, pblID)

  | mstep'2mstep m = 
  raise error ("mstep'2mstep: not impl. for "^(mstep'2str m));




(** decompose mstep' to a rule and to (lhs,rhs)
    unly needed                            ~~~ **)

val idT = Type ("Script.ID",[]);
(*val tt = (term_of o the o (parse thy)) "square_equation_left::ID";
type_of tt = idT;
val it = true : bool
*)
(* 13.3.01
v
*)
fun make_rule thy t =
  let val ct = cterm_of (sign_of thy) (Trueprop $ t)
  in Thm (string_of_cterm ct, make_thm ct) end;

(* val (Rewrite_Inst'(thy',rod,rls,put,subs,(thmID,thm),f,(f',asm)))=m;
   *)
(*decompose mstep' to a rule and to (lhs,rhs) for ets*)
fun rep_mstep' (Rewrite_Inst' 
		 (thy',rod,rls,put,subs,(thmID,thm),f,(f',asm))) = 
  let val fT = type_of f;
    val b = if put then HOLogic.true_const else HOLogic.false_const;
    val sT = (type_of o fst o hd) subs;
    val subs' = list2isalist (HOLogic.mk_prodT (sT, sT))
      (map HOLogic.mk_prod subs);
    val sT' = type_of subs';
    val lhs = Const ("Script.Rewrite'_Inst",[sT',idT,(*fT*)bool,fT] ---> fT) 
      $ subs' $ Free (thmID,idT) $ b $ f;
  in (((make_rule (assoc_thy thy')) o HOLogic.mk_eq) (lhs,f'),(lhs,f')) end
(*Fehlersuche 25.4.01
(a)----- als String zusammensetzen:
ML> Sign.string_of_term (sign_of thy)f; 
val it = "d_d x #4 + d_d x (x ^^^ #2 + #3 * x)" : string
ML> Sign.string_of_term (sign_of thy)f'; 
val it = "#0 + d_d x (x ^^^ #2 + #3 * x)" : string
ML> subs;
val it = [(Free ("bdv","RealDef.real"),Free ("x","RealDef.real"))] : subst
> val tt = (term_of o the o (parse thy))
  "(Rewrite_Inst[(bdv,x)]diff_const False(d_d x #4 + d_d x (x ^^^ #2 + #3 * x)))=(#0 + d_d x (x ^^^ #2 + #3 * x))";
> atomty thy tt;
ML> writeln(Sign.string_of_term (sign_of thy)tt); 
(Rewrite_Inst [(bdv,x)] diff_const False d_d x #4 + d_d x (x ^^^ #2 + #3 * x)) =
 #0 + d_d x (x ^^^ #2 + #3 * x)

(b)----- laut rep_mstep':
> val ttt=HOLogic.mk_eq (lhs,f');
> atomty thy ttt;


(*Fehlersuche 1-2Monate vor 4.01:*)
> val tt = (term_of o the o (parse thy))
  "Rewrite_Inst[(bdv,x)]square_equation_left True(x=#1+#2)";
> atomty thy tt;

> val f = (term_of o the o (parse thy)) "x=#1+#2";
> val f' = (term_of o the o (parse thy)) "x=#3";
> val subs = [((term_of o the o (parse thy)) "bdv",
	       (term_of o the o (parse thy)) "x")];
> val sT = (type_of o fst o hd) subs;
> val subs' = list2isalist (HOLogic.mk_prodT (sT, sT))
			      (map HOLogic.mk_prod subs);
> val sT' = type_of subs';
> val lhs = Const ("Script.Rewrite'_Inst",[sT',idT,fT,fT] ---> fT) 
  $ subs' $ Free (thmID,idT) $ HOLogic.true_const $ f;
> lhs = tt;
val it = true : bool
> rep_mstep' (Rewrite_Inst' 
	       ("Script.thy","tless_true","eval_rls",false,subs,
		("square_equation_left",""),f,(f',[])));
*)
  | rep_mstep' (Rewrite' (thy',rod,rls,put,(thmID,thm),f,(f',asm)))=
  let 
    val fT = type_of f;
    val b = if put then HOLogic.true_const else HOLogic.false_const;
    val lhs = Const ("Script.Rewrite",[idT,HOLogic.boolT,fT] ---> fT)
      $ Free (thmID,idT) $ b $ f;
  in (((make_rule (assoc_thy thy')) o HOLogic.mk_eq) (lhs,f'),(lhs,f')) end
(* 
> val tt = (term_of o the o (parse thy)) (*____   ____..test*)
  "Rewrite square_equation_left True (x=#1+#2) = (x=#3)";

> val f = (term_of o the o (parse thy)) "x=#1+#2";
> val f' = (term_of o the o (parse thy)) "x=#3";
> val Thm (id,thm) = 
  rep_mstep' (Rewrite' 
   ("Script.thy","tless_true","eval_rls",false,
    ("square_equation_left",""),f,(f',[])));
> val Some ct = parse thy   
  "Rewrite square_equation_left True (x=#1+#2)"; 
> rewrite_ Script.thy tless_true eval_rls true thm ct;
val it = Some ("x = #3",[]) : (cterm * cterm list) option
*)
  | rep_mstep' (Rewrite_Set_Inst' 
		 (thy',put,subs,rls,f,(f',asm))) = 
  let val fT = type_of f;
    val sT = (type_of o fst o hd) subs;
    val subs' = list2isalist (HOLogic.mk_prodT (sT, sT))
      (map HOLogic.mk_prod subs);
    val sT' = type_of subs';
    val b = if put then HOLogic.true_const else HOLogic.false_const
    val lhs = Const ("Script.Rewrite'_Set'_Inst",
		     [sT',idT,fT,fT] ---> fT) 
      $ subs' $ Free (id_rls rls,idT) $ b $ f;
  in (((make_rule (assoc_thy thy')) o HOLogic.mk_eq) (lhs,f'),(lhs,f')) end
(* ... vals from Rewrite_Inst' ...
> rep_mstep' (Rewrite_Set_Inst' 
	       ("Script.thy",false,subs,
		"isolate_bdv",f,(f',[])));
*)
(* val (Rewrite_Set' (thy',put,rls,f,(f',asm)))=m;
*)
  | rep_mstep' (Rewrite_Set' (thy',put,rls,f,(f',asm)))=
  let val fT = type_of f;
    val b = if put then HOLogic.true_const else HOLogic.false_const;
    val lhs = Const ("Script.Rewrite'_Set",[idT,bool,fT] ---> fT) 
      $ Free (id_rls rls,idT) $ b $ f;
  in (((make_rule (assoc_thy thy')) o HOLogic.mk_eq) (lhs,f'),(lhs,f')) end
(* 13.3.01:
val thy = assoc_thy thy';
val t = HOLogic.mk_eq (lhs,f');
make_rule thy t;
--------------------------------------------------
val lll = (term_of o the o (parse thy)) 
  "Rewrite_Set SqRoot_simplify False (d_d x (x ^^^ #2 + #3 * x) + d_d x #4)";

--------------------------------------------------
> val f = (term_of o the o (parse thy)) "x=#1+#2";
> val f' = (term_of o the o (parse thy)) "x=#3";
> val Thm (id,thm) = 
  rep_mstep' (Rewrite_Set' 
   ("Script.thy",false,"SqRoot_simplify",f,(f',[])));
val id = "(Rewrite_Set SqRoot_simplify True x = #1 + #2) = (x = #3)" : string
val thm = "(Rewrite_Set SqRoot_simplify True x = #1 + #2) = (x = #3)" : thm
*)
  | rep_mstep' (Calculate' (thy',op_,f,(f',thm')))=
  let val fT = type_of f;
    val lhs = Const ("Script.Calculate",[idT,fT] ---> fT) 
      $ Free (op_,idT) $ f
  in (((make_rule (assoc_thy thy')) o HOLogic.mk_eq) (lhs,f'),(lhs,f')) end
(*
> val lhs'=(term_of o the o (parse thy))"Calculate plus (#1+#2)";
  ... test-root-equ.sml: calculate ...
> val Appl m'=applicable_in p pt (Calculate "plus");
> val (lhs,_)=mstep'2etac m';
> lhs'=lhs;
val it = true : bool*)
  | rep_mstep' m = raise error ("rep_mstep': not impl.for "^
				 (mstep'2str m));


fun mstep'2rule m = (fst o rep_mstep') m;
fun mstep'2etac m = (snd o rep_mstep') m;
fun mstep'2tac m = (fst o snd o rep_mstep') m;
fun mstep'2res m = (snd o snd o rep_mstep') m;(*ONLYuse of rep_mstep'
					        FIXXXXME: simplify rep_mstep'*)


(* use"ME/script.sml";
   use"script.sml";
   *)




(** locate an applicable tactic in a script **)

datatype assoc = (*ExprVal in the sense of denotational semantics*)
  Assoc of     (*the tac is associated, strongly or weakly*)
  scrstate *       (*the current; returned for next_tac etc. outside ass* *)  
  (step list)    (*list of steps done until associated tac found;
	           initiated with the data for doing the 1st step,
                   thus the head holds these data further on,
		   while the tail holds steps finished (incl.scrstate in ptree)*)
| NasApp of   (*tac not associated, but applicable, ptree-node generated*)
  scrstate * (step list)
| NasNap of     (*tac not associated, not applicable, nothing generated;
	         for distinction in Or, for leaving iterations, leaving Seq,
		 evaluate scriptexpressions*)
  term * env;
fun assoc2str (Assoc     _) = "Assoc"
  | assoc2str (NasNap  _) = "NasNap"
  | assoc2str (NasApp _) = "NasApp";


datatype asap = (*arg. of assy _only_ for distinction w.r.t. Or*)
  Aundef   (*undefined: set only by (topmost) Or*)
| AssOnly  (*do not execute appl tacs - there could be an associated
	     in parallel Or-branch*)
| AssGen;  (*no Ass(Weak) found within Or, thus 
             search for _applicable_ tacs, execute and generate pt*)
(*this constructions doesnt allow arbitrary nesting of Or !!!*)


(*assy, ass_up, astep_up scanning for locate_gen at tactic in a script.
  search is clearly separated into (1)-(2):
  (1) assy is recursive descent;
  (2) ass_up resumes interpretation at a location somewhere in the script;
      astep_up does only get to the parentnode of the scriptexpr.
  consequence:
  * call of (2) means _always_ that in this branch below
    there was an appl.tac (Repeat, Or e1, ...)
*)

fun assy ya ((E,l,a,v,S,b),ss)
	  (Const ("Let",_) $ e $ (Abs (id,T,body))) =
    (case assy ya ((E , l@[L,R], a,v,S,b),ss) e of
	 NasApp ((E',l,a,v,S,bb),ss) => 
	 let val id' = mk_Free (id, T);
	     val E' = upd_env E' (id', v);
	 (*val _=writeln("### assy Let -> NasApp");*)
	 in assy ya ((E', l@[R,D], a,v,S,b),ss) body end
     | NasNap (v,E) => 	 
	 let val id' = mk_Free (id, T);
	   val E' = upd_env E (id', v);
	   (*val _=writeln("### assy Let -> NasNap");*)
	 in assy ya ((E', l@[R,D], a,v,S,b),ss) body end
     | ay => ay)

  | assy (ya as (((thy,srls),_),_)) ((E,l,_,v,S,b),ss) 
	 (Const ("Script.While",_) $ c $ e $ a) =
    ((*writeln("### assy While $ c $ e $ a, upd_env= "^
	     (subst2str (upd_env E (a,v))));*)
     if eval_true_ thy srls (subst_atomic (upd_env E (a,v)) c) 
     then assy ya ((E, l@[L,R], Some a,v,S,b),ss)  e
     else NasNap (v, E))
   
  | assy (ya as (((thy,srls),_),_)) ((E,l,a,v,S,b),ss) 
	 (Const ("Script.While",_) $ c $ e) =
    ((*writeln("### assy While, l= "^(loc_2str l));*)
     if eval_true_ thy srls (subst_atomic (upd_env_opt E (a,v)) c) 
     then assy ya ((E, l@[R], a,v,S,b),ss) e
     else NasNap (v, E)) 

  | assy (ya as (((thy,srls),_),_)) ((E,l,a,v,S,b),ss) 
	 (Const ("If",_) $ c $ e1 $ e2) =
    (if eval_true_ thy srls (subst_atomic (upd_env_opt E (a,v)) c) 
     then assy ya ((E, l@[L,R], a,v,S,b),ss) e1
     else assy ya ((E, l@[  R], a,v,S,b),ss) e2) 

  | assy ya ((E,l,_,v,S,b),ss) (Const ("Script.Try",_) $ e $ a) =
  ((*writeln("### assy Try, l= "^(loc_2str l));*)
    case assy ya ((E, l@[L,R], Some a,v,S,b),ss) e of
     ay => ay) 

  | assy ya ((E,l,a,v,S,b),ss) (Const ("Script.Try",_) $ e) =
  ((*writeln("### assy Try, l= "^(loc_2str l));*)
    case assy ya ((E, l@[R], a,v,S,b),ss) e of
     ay => ay)

  | assy ya ((E,l,_,v,S,b),ss) (Const ("Script.Seq",_) $e1 $ e2 $ a) =
    ((*writeln("### assy Seq $e1 $ e2 $ a, E= "^(subst2str E));*)
     case assy ya ((E, l@[L,L,R], Some a,v,S,b),ss) e1 of
	 NasNap (v, E) => assy ya ((E, l@[L,R], Some a,v,S,b),ss) e2
       | NasApp ((E,_,_,v,_,_),ss) => 
	 assy ya ((E, l@[L,R], Some a,v,S,b),ss) e2
       | ay => ay)

  | assy ya ((E,l,a,v,S,b),ss) (Const ("Script.Seq",_) $e1 $ e2) =
    (case assy ya ((E, l@[L,R], a,v,S,b),ss) e1 of
	 NasNap (v, E) => assy ya ((E, l@[R], a,v,S,b),ss) e2
       | NasApp ((E,_,_,v,_,_),ss) => 
	 assy ya ((E, l@[R], a,v,S,b),ss) e2
       | ay => ay)
    
  | assy ya ((E,l,_,v,S,b),ss) (Const ("Script.Repeat",_) $ e $ a) =
    assy ya ((E,(l@[L,R]),Some a,v,S,b),ss) e

  | assy ya ((E,l,a,v,S,b),ss) (Const ("Script.Repeat",_) $ e) =
    assy ya ((E,(l@[R]),a,v,S,b),ss) e

(*15.6.02: ass,app Or nochmals "uberlegen FIXXXME*)
  | assy (y, Aundef) ((E,l,_,v,S,b),ss) (Const ("Script.Or",_) $e1 $ e2 $ a) =
    (case assy (y, AssOnly) ((E,(l@[L,L,R]),Some a,v,S,b),ss) e1 of
	 NasNap (v, E) => 
	 (case assy (y, AssOnly) ((E,(l@[L,R]),Some a,v,S,b),ss) e2 of
	      NasNap (v, E) => 
	      (case assy (y, AssGen) ((E,(l@[L,L,R]),Some a,v,S,b),ss) e1 of
	       NasNap (v, E) => 
	       assy (y, AssGen) ((E, (l@[L,R]), Some a,v,S,b),ss) e2
	     | ay => ay)
	    | ay =>(ay))
       | NasApp _ => raise error ("assy: FIXXXME ///must not return NasApp///")
       | ay => (ay))

  | assy ya ((E,l,a,v,S,b),ss) (Const ("Script.Or",_) $e1 $ e2) =
    (case assy ya ((E,(l@[L,R]),a,v,S,b),ss) e1 of
	 NasNap (v, E) => 
	 assy ya ((E,(l@[R]),a,v,S,b),ss) e2
       | ay => (ay)) 
(* val ((m,_,pt,(p,p_),c)::ss) = [(m,EmptyMout,pt,p,[])];
   val t = (term_of o the o (parse Isac.thy)) "Rewrite rmult_1 False";
   *) 

  | assy (((thy',srls),d),ap) ((E,l,a,v,S,b), (m,_,pt,(p,p_),c)::ss) t =
  if is_listexpr t
    then (writeln"### assy: code listexpr missing"; NasNap (v, E))
  else let val p' = case p_ of Frm => p | Res => lev_on p
	| _ => raise error ("assy: call by "^(pos'2str (p,p_)));
      val tac = subst_tac E a v t;
  in case assod d m tac of
	 Ass (m,v') =>
	 let val (p'',c',f',pt') = generate1 (assoc_thy thy') m 
			        (ScrState (E,l,a,v',S,true)) (p',p_) pt;
	   in Assoc ((E,l,a,v',S,true), (m,f',pt',p'',c @ c')::ss) end
       | AssWeak (m,v') => 
	   let val (p'',c',f',pt') = generate1 (assoc_thy thy') m 
			         (ScrState (E,l,a,v',S,false)) (p',p_) pt;
	   in Assoc ((E,l,a,v',S,false), (m,f',pt',p'',c @ c')::ss) end
       | NotAss =>
	   (case ap of   (*switch for Or: 1st AssOnly, 2nd AssGen*)
	      AssOnly => (NasNap (v, E))
	    | gen => (case applicable_in (p,p_) pt 
					 (tac2mstep (assoc_thy thy') tac) of
			Appl m' =>
			  let val is = (E,l,a,mstep'2res m',S,false(*FIXXXME*))
			      val (p'',c',f',pt') =
			      generate1 (assoc_thy thy') m' (ScrState is) (p',p_) pt;
			  in NasApp (is,(m,f',pt',p'',c @ c')::ss) end
		      | Notappl _ => 
			    (NasNap (v, E))
			    )
		)
       end;
(* (astep_up ((thy',scr,d),NasApp_) ((E,l,a,v,S,b),[(m,EmptyMout,pt,p,[])])) handle e => print_exn_G e;
  *)

fun ass_up (ys as (y,s,Script sc,d)) ((E,l,a,v,S,b),ss) 
	   (Const ("Let",_) $ _) =
    let val l = drop_last l; (*comes from e, goes to Abs*)
      val (Const ("Let",_) $ e $ (Abs (i,T,body))) = go l sc;
      val i = mk_Free (i, T);
      val E = upd_env E (i, v);
      (*val _=writeln("### ass_up Let e: E="^(subst2str E));*)
    in case assy (((y,s),d),Aundef) ((E, l@[R,D], a,v,S,b),ss) body of
	   Assoc iss => Assoc iss
	 | NasApp iss => astep_up ys iss 
	 | NasNap (v, E) => astep_up ys ((E,l,a,v,S,b),ss) end

  | ass_up ys iss (Abs (_,_,_)) = 
    astep_up ys iss (*TODO 5.9.00: env ?*)

  | ass_up ys iss (Const ("Let",_) $ e $ (Abs (i,T,b))) =
    astep_up ys iss (*TODO 5.9.00: env ?*)


  | ass_up ysa iss (Const ("Script.Seq",_) $ _ $ _ $ _) = 
    astep_up ysa iss (*all has been done in (*2*) below*)

  | ass_up ysa iss (Const ("Script.Seq",_) $ _ $ _) = 
    astep_up ysa iss (*2*: comes from e2*)

  | ass_up (ysa as (y,s,Script sc,d)) ((E,l,a,v,S,b),ss)
	   (Const ("Script.Seq",_) $ _ ) = (*2*: comes from e1, goes to e2*)
    let val up = drop_last l;
	val Const ("Script.Seq",_) $ _ $ e2 = go up sc;
    in case assy (((y,s),d),Aundef) ((E, up@[R], a,v,S,b),ss) e2 of
	   NasNap (v,E) => astep_up ysa ((E,up,a,v,S,b),ss)
	 | NasApp iss => astep_up ysa iss
	 | ay => ay end

  | ass_up ysa iss (Const ("Script.Try",_) $ e $ _) =
    astep_up ysa iss

  | ass_up ysa iss (Const ("Script.Try",_) $ e) =
    astep_up ysa iss

  | ass_up (ys as (y,s,_,d)) ((E,l,_,v,S,b),ss)
	   (Const ("Script.While",_) $ c $ e $ a) =
    ((*writeln("### ass_up: While c= "^
	     (term2str (subst_atomic (upd_env E (a,v)) c)));*)
     if eval_true_ y s (subst_atomic (upd_env E (a,v)) c)
    then (case assy (((y,s),d),Aundef) ((E, l@[L,R], Some a,v,S,b),ss) e of 
       NasNap (v,E') => astep_up ys ((E',l, Some a,v,S,b),ss)
     | NasApp ((E',l,a,v,S,b),ss) =>
       ass_up ys ((E',l,a,v,S,b),ss) t
     | ay => ay)
    else astep_up ys ((E,l, Some a,v,S,b),ss)
	 )

  | ass_up (ys as (y,s,_,d)) ((E,l,a,v,S,b),ss)
	   (Const ("Script.While",_) $ c $ e) =
    if eval_true_ y s (subst_atomic (upd_env_opt E (a,v)) c)
    then (case assy (((y,s),d),Aundef) ((E, l@[R], a,v,S,b),ss) e of 
       NasNap (v,E') => astep_up ys ((E',l, a,v,S,b),ss)
     | NasApp ((E',l,a,v,S,b),ss) =>
       ass_up ys ((E',l,a,v,S,b),ss) t
     | ay => ay)
    else astep_up ys ((E,l, a,v,S,b),ss)

  | ass_up y iss (Const ("If",_) $ _ $ _ $ _) = astep_up y iss

  | ass_up (ys as (y,s,_,d)) ((E,l,_,v,S,b),ss)
	   (t as Const ("Script.Repeat",_) $ e $ a) =
  (case assy (((y,s),d), Aundef) ((E, (l@[L,R]), Some a,v,S,b),ss) e of 
       NasNap (v,E') => astep_up ys ((E',l, Some a,v,S,b),ss)
     | NasApp ((E',l,a,v,S,b),ss) =>
       ass_up ys ((E',l,a,v,S,b),ss) t
     | ay => ay)

  | ass_up (ys as (y,s,_,d)) (is as ((E,l,a,v,S,b),ss)) 
	   (t as Const ("Script.Repeat",_) $ e) =
  (case assy (((y,s),d), Aundef) ((E, (l@[R]), a,v,S,b),ss) e of 
       NasNap (v', E') => astep_up ys ((E',l,a,v',S,b),ss)
     | NasApp ((E',l,a,v',S,bb),ss) => 
       ass_up ys ((E',l,a,v',S,b),ss) t
     | ay => ay)

  | ass_up y iss (Const ("Script.Or",_) $ _ $ _ $ _) = astep_up y iss

  | ass_up y iss (Const ("Script.Or",_) $ _ $ _) = astep_up y iss

  | ass_up y ((E,l,a,v,S,b),ss) (Const ("Script.Or",_) $ _ ) = 
    astep_up y ((E, (drop_last l), a,v,S,b),ss)

  | ass_up  y iss t =
    raise error ("ass_up not impl for t= "^(term2str t))
	  
and astep_up (ys as (_,_,Script sc,_)) ((E,l,a,v,S,b),ss) =
  if 1 < length l 
    then 
      let val up = drop_last l;
	  (*val _= writeln("### astep_up: v= "^(term2str v));*)
      in ass_up ys ((E,up,a,v,S,b),ss) (go up sc) end
  else (NasNap (v, E))
;





(* use"ME/script.sml";
   use"script.sml";
   *)

(*check if there are msteps for rewriting only*)
fun rew_only ([]:step list) = true
  | rew_only (((Rewrite' _          ,_,_,_,_))::ss) = rew_only ss
  | rew_only (((Rewrite_Inst' _     ,_,_,_,_))::ss) = rew_only ss
  | rew_only (((Rewrite_Set' _      ,_,_,_,_))::ss) = rew_only ss
  | rew_only (((Rewrite_Set_Inst' _ ,_,_,_,_))::ss) = rew_only ss
  | rew_only (((Calculate' _        ,_,_,_,_))::ss) = rew_only ss
  | rew_only (((Begin_Trans' _      ,_,_,_,_))::ss) = rew_only ss
  | rew_only (((End_Trans' _        ,_,_,_,_))::ss) = rew_only ss
  | rew_only _ = false; 
  

datatype locate =
  Steps of istate      (*producing hd of step list (which was latest)
	                 for next_tac, for reporting Safe|Unsafe to DG*)
	   * step      (*(scrstate producing this step is in ptree !)*) 
		 list  (*locate_gen may produce intermediate steps*)
| NotLocatable;        (*no (m Ass m') or (m AssWeak m') found*)



(* locate_gen tries to locate an input mstep m in the script. 
   pursuing this goal the script is executed until an (m' equiv m) is found,
   or the end of the script
args
   m   : input by the user, already checked by applicable_in,
         (to be searched within Or; and _not_ an m doing the step on ptree !)
   p,pt: (incl ets) at the time of input
   scr : the script
   d   : canonical simplifier for locating Take, Substitute, Subproblems etc.
   ets : ets at the time of input
   l   : the location (in scr) of the tac which generated the current formula
returns
   Steps: pt,p (incl. ets) with m done
          pos' list of proofobjs cut (from generate)
          safe: implied from last proofobj
	  ets:
   ///ToDo : ets contains a list of msteps to be done before m can be done
          NOT IMPL. -- "error: do other step before"
   NotLocatable: thus generate_hard
*)
(* val (Rewrite'(_,ro,er,pa,(id,str),f,_), p, Rfuns {locate_rule=lo,...},
	RrlsState (_,f'',rss,rts)) = (m, (p,p_), sc, is);
   *)
fun locate_gen (thy',_) (Rewrite'(_,ro,er,pa,(id,str),f,_)) (pt,p) 
	       (Rfuns {locate_rule=lo,...}, d) (RrlsState (_,f'',rss,rts)) = 
    (case lo rss f (Thm (id, mk_thm (assoc_thy thy') str)) of
	 [] => NotLocatable
       | rts' => 
	 Steps (rts2steps [] ((pt,p),(f,f'',rss,rts),(thy',ro,er,pa)) rts'))
(* val p as(p',p_)=(p,p_);val scr as Script(h $ body)=sc;val (E,l,a,v,S,bb)=is;
   locate_gen (thy':theory') (m:mstep') ((pt,p):ptree * pos') 
	      (scr,d) (E,l,a,v,S,bb);

   val (p) = (p,p_);
   val (scr as Script (h $ body)) = (sc);
   val ScrState (E,l,a,v,S,b) = (is);
   locate_gen thy' m (pt,p) (Script (h $ body),d) (ScrState(E,l,a,v,S,b));
   *)
  | locate_gen (ts as (thy',srls)) (m:mstep') ((pt,p):ptree * pos') 
	       (scr as Script (h $ body),d) (ScrState (E,l,a,v,S,b))  = 
  let (*val _= writeln("### locate_gen: p="^(pos'2str p));*)
      val thy = assoc_thy thy';
  in case if l=[] then (assy ((ts,d),Aundef) ((E,[R],a,v,S,b),
						[(m,EmptyMout,pt,p,[])]) body)
	  else (astep_up (thy',srls,scr,d) ((E,l,a,v,S,b),
						[(m,EmptyMout,pt,p,[])])  ) of
	 Assoc (iss as (is as (_,_,_,_,_,bb), ss as ((m',f',pt',p',c')::_))) =>
	 ((*writeln("### locate_gen: p'="^(pos'2str p'));*)
	  if bb then Steps (ScrState is, ss)
	 else if rew_only ss (*andalso 'not bb'= associated weakly*)
	 then let (*val _=writeln("### locate_gen, bef g1: p="^(pos'2str p));*)
		  val (po,p_) = p;
                  val po' = case p_ of Frm => po | Res => lev_on po
                  val (p'',c'',f'',pt'') = generate1 thy m (ScrState is) (po',p_) pt;
	      (*val _=writeln("### locate_gen, aft g1: p''="^(pos'2str p''));*)
	      (*drop the intermediate steps !*)
	      in Steps (ScrState is, [(m, f'',pt'',p'',c'')]) end
	 else Steps (ScrState is, ss))
	
     | NasApp _ (*[((E,l,a,v,S,bb),(m',f',pt',p',c'))] => 
	   raise error ("locate_gen: should not have got NasApp, ets =")*)
       => NotLocatable
     | NasNap (_,_) =>
       if l=[] then NotLocatable
       else (*scan from begin of script for rew_only*)
	   (case assy ((ts,d),Aundef) ((E,[R],a,v,Unsafe,b),
					 [(m,EmptyMout,pt,p,[])]) body  of
		Assoc (iss as (is as (_,_,_,_,_,bb), 
			       ss as ((m',f',pt',p',c')::_))) =>
		    ((*writeln"4### locate_gen Assoc after Fini";*)
		     if rew_only ss
		     then let val(p'',c'',f'',pt'') = generate1 thy m (ScrState is) p' pt;
			  (*drop the intermediate steps !*)
			  in Steps (ScrState is, [(m, f'',pt'',p'',c'')]) end
		     else NotLocatable)
	      | _ => ((*writeln ("#### locate_gen: after Fini");*)
		      NotLocatable))
  end
  | locate_gen _ m _ (sc,_) is = 
    raise error ("locate_gen: wrong arguments,\n mstep= "^(mstep'2str m)^
		 ",\n scr= "^(scr2str sc)^",\n istate= "^(istate2str is));



(** find the next tactic in a script **)

datatype appy =  (*ExprVal in the sense of denotational semantics*)
    Appy of      (*applicable tac found, search stalled*)
    loc_ * mstep'  (*loc_ not yet used 28.4.02*)
  | Napp of      (*tac found was not applicable; 
	           this mode may become Skip in Repeat, Try and Or*)
    env (*stack*)  (*popped while nxt_up*)
  | Skip of      (*for restart after Appy, for leaving iterations,
	           for passing the value of scriptexpressions,
		   and for finishing the script successfully*)
    term * env (*stack*);

(*appy, nxt_up, nstep_up scanning for next_tac.
  search is clearly separated into (1)-(2):
  (1) appy is recursive descent;
  (2) nxt_up resumes interpretation at a location somewhere in the script;
      nstep_up does only get to the parentnode of the scriptexpr.
  consequence:
  * call of (2) means _always_ that in this branch below
    there was an appl.tac (Repeat, Or e1, ...)
*)

datatype appy_ = (*as argument in nxt_up, nstep_up, from appy*)
       (*  Appy is only (final) returnvalue, not argument during search
       |*) Napp_ (*ev. detects 'script is not appropriate for this example'*)
       | Skip_;  (*detects 'script successfully finished'
		   also used as init-value for resuming; this works,
	           because 'nxt_up Or e1' treats as Appy*)

fun appy thy ptp E l
  (t as Const ("Let",_) $ e $ (Abs (i,T,b))) a v =
  (case appy thy ptp E (l@[L,R]) e a v of
     Skip (res, E) => 
       let (*val _= writeln("### appy Let "^(term2str t));
	 val _= writeln("### appy Let: Skip res ="^(term2str res));*)
	 val (i',b') = variant_abs (i,T,b);    (*term.ML FIXXXXXME streichen?*)
	 val i = mk_Free(i',T);
	 val E' = upd_env E (i,res);
       in appy thy ptp E' (*eno,[](*env'*),iar,res,s*) (l@[R,D]) b(*ody'*) a v end
   | ay => ay)

  | appy (thy as (th,sr)) ptp E l
  (t as Const ("Script.While"(*1*),_) $ c $ e $ a) _ v = (*ohne n. 28.9.00*)
  ((*writeln("### appy While $ c $ e $ a, upd_env= "^
	   (subst2str (upd_env E (a,v))));*)
   if eval_true_ th sr (subst_atomic (upd_env E (a,v)) c)
    then appy thy ptp E (l@[L,R]) e (Some a) v
  else Skip (v, E))

  | appy (thy as (th,sr)) ptp E l
  (t as Const ("Script.While"(*2*),_) $ c $ e) a v =(*ohne nachdenken 28.9.00*)
  ((*writeln("### appy While $ c $ e, upd_env= "^
	   (subst2str (upd_env_opt E (a,v))));*)
   if eval_true_ th sr (subst_atomic (upd_env_opt E (a,v)) c)
    then appy thy ptp E (l@[R]) e a v
  else Skip (v, E))

  | appy (thy as (th,sr)) ptp E l (t as Const ("If",_) $ c $ e1 $ e2) a v =
    ((*writeln("### appy If: t= "^(term2str t));
     writeln("### appy If: c= "^(term2str (subst_atomic (upd_env_opt E (a,v))c)));
     writeln("### appy If: thy= "^thy);*)
     if eval_true_ th sr (subst_atomic (upd_env_opt E (a,v)) c)
     then ((*writeln("### appy If: true") ;*)appy thy ptp E (l@[L,R]) e1 a v)
     else ((*writeln("### appy If: false");*)appy thy ptp E (l@[  R]) e2 a v))

  | appy thy ptp E (*env*) l
  (Const ("Script.Repeat"(*1*),_) $ e $ a) _ v = 
    ((*writeln("### appy Repeat a: ");*)
     appy thy ptp E (*env*) (l@[L,R]) e (Some a) v)

  | appy thy ptp E (*env*) l
  (Const ("Script.Repeat"(*2*),_) $ e) a v = 
    ((*writeln("3### appy Repeat: a= "^
	     (Sign.string_of_term (sign_of (assoc_thy thy)) a));*)
     appy thy ptp E (*env*) (l@[R]) e a v)

  | appy thy ptp E l
  (t as Const ("Script.Try",_) $ e $ a) _ v =
  (case appy thy ptp E (l@[L,R]) e (Some a) v of
     Napp E => ((*writeln("### appy Try "^
			  (Sign.string_of_term (sign_of (assoc_thy thy)) t));*)
		 Skip (v, E))
   | ay => ay)
(* val (l,t as Const ("Script.Try",_) $ e) = (l@[L,R],e1);
   appy thy ptp E l t a v;
   *)
  | appy thy ptp E l
  (t as Const ("Script.Try",_) $ e) a v =
  (case appy thy ptp E (l@[R]) e a v of
     Napp E => ((*writeln("### appy Try "^
			  (Sign.string_of_term (sign_of (assoc_thy thy)) t));*)
		 Skip (v, E))
   | ay => ay)


  | appy thy ptp E l
	 (Const ("Script.Or"(*1*),_) $e1 $ e2 $ a) _ v =
    (case appy thy ptp E (l@[L,L,R]) e1 (Some a) v of
	 Appy lm => Appy lm
       | _ => appy thy ptp E (*env*) (l@[L,R]) e2 (Some a) v)
    
  | appy thy ptp E l
	 (Const ("Script.Or"(*2*),_) $e1 $ e2) a v =
    (case appy thy ptp E (l@[L,R]) e1 a v of
	 Appy lm => Appy lm
       | _ => appy thy ptp E (l@[R]) e2 a v)
    

  | appy thy ptp E l
	 (Const ("Script.Seq"(*1*),_) $e1 $ e2 $ a) _ v =
    ((*writeln("### appy Seq $ e1 $ e2 $ a, upd_env= "^
	     (subst2str (upd_env E (a,v))));*)
     case appy thy ptp E (l@[L,L,R]) e1 (Some a) v of
	 Skip (v,E) => appy thy ptp E (l@[L,R]) e2 (Some a) v
       | ay => ay)
(* val (l,t as Const ("Script.Seq",_) $e1 $ e2) = (up@[R],e2);
   appy thy ptp E l t a v;
   *)    
  | appy thy ptp E l
	 (Const ("Script.Seq",_) $e1 $ e2) a v =
    (case appy thy ptp E (l@[L,R]) e1 a v of
	 Skip (v,E) => appy thy ptp E (l@[R]) e2 a v
       | ay => ay)
    
  (*here at the end, this must be a tactic
    wrt. diss. missing checks: tac in ets, tac ready*)
(* val (pt, p) = ptp;
   val t = (term_of o the o (parse SqRoot.thy))
	       "Rewrite square_equation_left True";

   val (pt, p) = ptp;
   val t = (term_of o the o (parseold SqRoot.thy))
	       "Check_elementwise L_ {v_. Assumptions} ";

   val ((pt,p),l,t,a,v) = (ptp,[R],body,None,e_term);
   val t = body;   
   val t = (term_of o the o (parseold SqRoot.thy))
	       "Rewrite square_equation_left True";
   val a = Some ((term_of o the o (parseold SqRoot.thy)) "e_::bool");

   val ((pt,p),l,t,a,v) = (ptp,[R],body,None,e_term);

   val (thy as (th,sr),(pt,p),l,t) = (thy,ptp,l@[R],e);
   appy (th,sr) (pt, p) E l t a v;
   *)
  | appy (thy as (th,sr)) (pt, p) E l t a v =
  if is_listexpr t
    then ((*writeln("### appy, listexpr= "^(term2str t));
	  writeln("### appy, eval(..)= "^(term2str
	    (eval_listexpr_ (assoc_thy thy) 
			    (subst_atomic (upd_env_opt E (a,v)) t))));*)
	  Skip (eval_listexpr_ (assoc_thy th) sr
			       (subst_atomic (upd_env_opt E (a,v)) t), E))
  else let val tac = subst_tac E a v t;
	 (*val _= writeln("### appy t, E= "^(subst2str E));
val t =(term_of o the o (parse Isac.thy))"Rewrite_Set make_polyomial False g_";
val v =(term_of o the o (parse Isac.thy))"x+1=0";
           val _= writeln("### appy t, a= "^(termopt2str a));
           val _= writeln("### appy t, v= "^(term2str v));
           val _= writeln("### appy t, tac= "^(term2str tac));
           val _= writeln("### appy t, vor  tac2mstep"); 
           val _= writeln("### appy t, l  = "^(loc_2str l));*)
	   val (m,_) = tac2mstep' (assoc_thy th) tac;
       (*val _= writeln("### appy t, nach tac2mstep, m= "^(mstep2str m));*)
       in (case applicable_in p pt m of
	     Appl m' => ((*writeln("### appy: Appy");*)Appy (l, m'))
	   | _ => ((*writeln("### appy: Napp");*)Napp E)) end;
	 

fun nxt_up thy ptp (scr as (Script sc)) E l ay
    (t as Const ("Let",_) $ _) a v = (*comes from let=...*)
    if ay = Napp_
    then nstep_up thy ptp scr E (drop_last l) Napp_ a v
    else (*Skip_*)
	let val up = drop_last l;
		val (Const ("Let",_) $ e $ (Abs (i,T,body))) = go up sc;
	in case appy thy ptp (*upd_env*) E (*a,v)*) (up@[R,D]) body a v  of
	       Appy lr => Appy lr
	     | Napp E => nstep_up thy ptp scr E up Napp_ a v
	     | Skip (v,E) => nstep_up thy ptp scr E up Skip_ a v end
	    
  | nxt_up thy ptp scr E l ay
    (t as Abs (_,_,_)) a v = 
    ((*writeln("### nxt_up Abs: "^
	     (Sign.string_of_term (sign_of (assoc_thy thy)) t));*)
     nstep_up thy ptp scr E (*enr*) l ay a v)

  | nxt_up thy ptp scr E l ay
    (t as Const ("Let",_) $ e $ (Abs (i,T,b))) a v =
    ((*writeln("### nxt_up Let e Abs: "^
	     (Sign.string_of_term (sign_of (assoc_thy thy)) t));*)
     nstep_up thy ptp scr (*upd_env*) E (*a,v)*) 
	      (*eno,upd_env env (iar,res),iar,res,saf*) l ay a v)

  (*no appy_: never causes Napp -> Helpless*)
  | nxt_up (thy as (th,sr)) ptp scr E l _ 
  (Const ("Script.While"(*1*),_) $ c $ e $ _) a v = 
  if eval_true_ th sr (subst_atomic (upd_env_opt E (a,v)) c) 
    then case appy thy ptp E (l@[L,R]) e a v of
	     Appy lr => Appy lr
	   | Napp E => nstep_up thy ptp scr E l Skip_ a v
	   | Skip (v,E) => nstep_up thy ptp scr E l Skip_ a v
  else nstep_up thy ptp scr E l Skip_ a v

  (*no appy_: never causes Napp - Helpless*)
  | nxt_up (thy as (th,sr)) ptp scr E l _ 
  (Const ("Script.While"(*2*),_) $ c $ e) a v = 
  if eval_true_ th sr (subst_atomic (upd_env_opt E (a,v)) c) 
    then case appy thy ptp E (l@[R]) e a v of
	     Appy lr => Appy lr
	   | Napp E => nstep_up thy ptp scr E l Skip_ a v
	   | Skip (v,E) => nstep_up thy ptp scr E l Skip_ a v
  else nstep_up thy ptp scr E l Skip_ a v

  | nxt_up thy ptp scr E l ay (Const ("If",_) $ _ $ _ $ _) a v = 
    nstep_up thy ptp scr E l ay a v

  | nxt_up thy ptp scr E l _ (*no appy_: there was already a tac below*)
  (Const ("Script.Repeat"(*1*),T) $ e $ _) a v =
    (case appy thy ptp (*upd_env*) E (*a,v)*) ((l@[L,R]):loc_) e a v  of
      Appy lr => Appy lr
    | Napp E => ((*writeln("### nxt_up Repeat a: ");*)
		 nstep_up thy ptp scr E l Skip_ a v)
    | Skip (v,E) => ((*writeln("### nxt_up Repeat: Skip res ="^
		(Sign.string_of_term(sign_of (assoc_thy thy)) res'));*)
		    nstep_up thy ptp scr E l Skip_ a v))

  | nxt_up thy ptp scr E l _ (*no appy_: there was already a tac below*)
  (Const ("Script.Repeat"(*2*),T) $ e) a v =
    (case appy thy ptp (*upd_env*) E (*a,v)*) ((l@[R]):loc_) e a v  of
      Appy lr => Appy lr
    | Napp E => ((*writeln("### nxt_up Repeat a: ");*)
		 nstep_up thy ptp scr E l Skip_ a v)
    | Skip (v,E) => ((*writeln("### nxt_up Repeat: Skip res ="^
		(Sign.string_of_term(sign_of (assoc_thy thy)) res'));*)
		    nstep_up thy ptp scr E l Skip_ a v))

(* val (scr,l,t) = (Script sc,up,go up sc);
   nxt_up thy ptp scr E l ay t a v;
   *)
  | nxt_up thy ptp scr E l _ (*makes Napp to Skip*)
  (t as Const ("Script.Try",_) $ e $ _) a v = 
    ((*writeln("### nxt_up Try "^
	     (Sign.string_of_term (sign_of (assoc_thy thy)) t));*)
     nstep_up thy ptp scr E l Skip_ a v )

  | nxt_up thy ptp scr E l _ (*makes Napp to Skip*)
  (t as Const ("Script.Try"(*2*),_) $ e) a v = 
    ((*writeln("### nxt_up Try "^
	     (Sign.string_of_term (sign_of (assoc_thy thy)) t));*)
     nstep_up thy ptp scr (*upd_env*) E (*a,v)*) l Skip_ a v)


  | nxt_up thy ptp scr E l ay
  (Const ("Script.Or",_) $ _ $ _ $ _) a v = nstep_up thy ptp scr E l ay a v

  | nxt_up thy ptp scr E l ay
  (Const ("Script.Or",_) $ _ $ _) a v = nstep_up thy ptp scr E l ay a v

  | nxt_up thy ptp scr E l ay
  (Const ("Script.Or",_) $ _ ) a v = 
    nstep_up thy ptp scr E (drop_last l) ay a v


  | nxt_up thy ptp scr E l ay (*all has been done in (*2*) below*)
  (Const ("Script.Seq"(*1*),_) $ _ $ _ $ _) a v =
    nstep_up thy ptp scr E l ay a v

  | nxt_up thy ptp scr E l ay (*comes from e2*)
	   (Const ("Script.Seq"(*2*),_) $ _ $ e2) a v =
    nstep_up thy ptp scr E l ay a v
(* val (scr as Script sc,l,t) = (Script sc,up,go up sc);
   !! Const ("Script.Seq",_ $ _ 
   nxt_up thy ptp scr E l ay t a v;
   *)
  | nxt_up thy ptp (scr as Script sc) E l ay (*comes from e1*)
	   (Const ("Script.Seq",_) $ _ ) a v = 
    if ay = Napp_
    then nstep_up thy ptp scr E (drop_last l) Napp_ a v
    else (*Skip_*)
	let val up = drop_last l;
	    val Const ("Script.Seq"(*2*),_) $ _ $ e2 = go up sc;
	in case appy thy ptp E (up@[R]) e2 a v  of
	    Appy lr => Appy lr
	  | Napp E => nstep_up thy ptp scr E up Napp_ a v
	  | Skip (v,E) => nstep_up thy ptp scr E up Skip_ a v end

  | nxt_up (thy,_) ptp scr E l ay t a v =
  raise error ("nxt_up not impl for "^
	       (Sign.string_of_term (sign_of (assoc_thy thy)) t))

(* val (Script sc, ay) = (sc,Skip_);
   val (Script sc, ay) = (scr,Skip_);
   *)
and nstep_up thy ptp (Script sc) E l ay a v = 
  ((*writeln("### nstep_up from: "^(loc_2str l));
   writeln("### nstep_up from: "^
	   (Sign.string_of_term (sign_of (assoc_thy thy)) (go l sc)));*)
   if 1 < length l 
   then 
       let val up = drop_last l; 
       in ((*writeln("### nstep_up to: "^
	      (Sign.string_of_term (sign_of (assoc_thy thy)) (go up sc)));*)
	   nxt_up thy ptp (Script sc) E up ay (go up sc) a v ) end
   else (*interpreted to end*)
       if ay = Skip_ then Skip (v, E) else Napp E 
);

(* decide for the next applicable tac in the script;
   returns (tactic, value) - the value in case the script is finished 
   12.8.02:         ~~~~~ and no assumptions ??? FIXME ???
   20.8.02: must return p in case of finished, because the next script
            consulted need not be the calling script:
            in case of detail ie. _inserted_ PrfObjs, the next tac
            has to searched in a script with PblObj.status<>Complete !
            (.. not true for other details ..PrfObj ??????????????????
   20.8.02: do NOT return safe (is only changed in locate !!!)
*)
(* val (thy, (pt,p), Rfuns {next_rule=ne,...}, RrlsState (f,f',rss,_)) = 
       (thy', (pt,p), sc, RrlsState (ii t));
   val (thy, (pt,p), Rfuns {next_rule=ne,...}, RrlsState (f,f',rss,_)) = 
       (thy', (pt',p'), sc, is');
   *)
fun next_tac thy (pt,p) (Rfuns {next_rule=ne,...}) (RrlsState (f,f',rss,_)) =
    if f = f' then (End_Detail, (f', Sundef(*FIXME is no value of next_tac!*)))
                                                          (*finished*)
    else (case ne rss f of
	      None => (Empty_Mstep, (e_term, Sundef)) 	  (*helpless*)
(* val Some (Thm (id,thm)) = ne rss f;
   *)
	    | Some (Thm (id,thm)) => (Rewrite (id, string_of_thm' thm), 
				    (e_term, Sundef)))    (*next tac*)
(* val (thy, ptp as (pt,(p,_)),sc as Script(h $ body)) =
       ((thy',srls),(pt,p), sc);
   
   val (thy, ptp as (pt,(p,_)),sc as Script(h $ body)) =
       ((thy',srls),(pt',p'), sc);
   val ScrState (E,l,a,v,s,b) = is';
   next_tac thy ptp (Script (h $ body)) (ScrState (E,l,a,v,s,b));
*)
  | next_tac thy (ptp as (pt,(p,_)):ptree * pos') (sc as Script (h $ body)) 
	     (ScrState (E,l,a,v,s,b)) =
  ((*writeln("### next_tac: p  = "^(ints2str p));
   writeln("### next_tac: ist= "^(istate2str (E,l,a,v,s,b)));
   writeln("### next_tac: = "^());*)
   case if l=[] then appy thy ptp E [R] body None e_term
       else nstep_up thy ptp sc E l Skip_ a v of
      Skip (v,_) =>                                        (*finished*)
      (case par_pbl_det pt p of
	   (true, p', _) => 
	   let val (_,pblID,_) = get_obj g_spec pt p';
	   in (Check_Postcond pblID, (v,s)) end
	 | (_,p',rls') => (End_Detail, (v,s)))             
    | Napp _ => (Empty_Mstep, (e_term, Sundef)) 	   (*helpless*)
    | Appy (_,m') => (mstep'2mstep m', (e_term, Sundef)))  (*next tac*)
  | next_tac _ _ _ is = raise error ("next_tac: not impl for "^
				     (istate2str is));

(* val metID = mI;
   *)
fun init_scrstate thy itms metID =
    let val ags = itms2args thy metID itms;
	val scr as Script sc = (#scr o get_met) metID;
        val env = ((actual_args sc) ~~ ags) 
	    handle _ => raise error ("init_scrstate: formal_args = "^
				   (terms2str (actual_args sc))^
				   "\ndon't match actual_args ="^
				   (terms2str ags));
    in (ScrState (env,[],None,e_term,Safe,true), scr):istate * scr end;
fun one_scr_arg (Const _ $ arg $ _) = arg
  | one_scr_arg t = raise error ("one_scr_arg: called by "^(term2str t));

(*.decide, where to get script and istate from:
   (*1*) from PblObj: script only
   (*2*) from PrfObj in case of detail a ruleset.*)
fun from_pblobj_or_detail' thy' (p,p_) pt = 
    let val (pbl,p',rls') = par_pbl_det pt p
    in if pbl 
       then (*1*)
	   let val thy = assoc_thy thy'
	       val PblObj{meth=itms,...} = get_obj I pt p'
	       val metID = get_obj g_metID pt p'
	       val {srls,...} = get_met metID
	   in if last_elem p = 0 (*nothing written to pt yet*)
	      then let val (is, sc) = init_scrstate thy itms metID
		   in (srls, is, sc) end
	      else (srls, get_istate pt (p,p_), (#scr o get_met) metID)
	   end
       else (*2*)
(* val Some rls = assoc(!ruleset', rls');
   *)
	   (*case assoc(!ruleset', rls') of !!!FIXME.3.4.03:re-organize!!!
	       None => raise error ("unknown ruleset '"^rls'^"'")
	     | Some rls =>*)(e_rls, (*unused for Rrls in locate_gen, next_tac*)
			    get_istate pt (p,p_),
			    case rls' of
				Rls {scr=scr,...} => scr
			      | Rrls {scr=rfuns,...} => rfuns)
    end;
(*.get script and istate from PblObj, see (*1*) above.*)
fun from_pblobj' thy' (p,p_) pt = 
    let val p' = par_pblobj pt p
	val thy = assoc_thy thy'
	val PblObj{meth=itms,...} = get_obj I pt p'
	val metID = get_obj g_metID pt p'
	val {srls,scr,...} = get_met metID
    in if last_elem p = 0 (*nothing written to pt yet*)
       then let val (is, scr) = init_scrstate thy itms metID
	    in (srls, is, scr) end
       else (srls, get_istate pt (p,p_), scr)
    end;
    
(*
end
open Interpreter;
*)

(* use"ME/script.sml";
   use"script.sml";
   *)
