(* global variables needed in ME, DG, and updated in knowledge
   WN.8.01 from modspec.sml
*)

(* data for methods stored in 'methods'-database *)
type meth' = 
     (*defaults for Rewrite' (rls, rrls have their own data):*)
     {rew_ord'  : rew_ord',   (*for rules in Detail*)
      erls      : rls,       (*the eval_rls for cond. in rules*)
      (*             '^ because of rewrite in applicable_in
						 compare type scr and rls*)
      srls      : rls,        (*for evaluating list expressions in scr*)
      prls      : rls,        (*for evaluating predicates in ppc*)
      calc      : calc list,  (*for Calculate in scr*)
      asm_thm   : thm' list,  (*put asm into environment*)
      asm_rls   : rls' list,  (*27.8.02: ??? unused (from default hard_gen ?)*)
      (* compare type pbt:*)
      ppc: ((string *      (*field*)
	     (term *       (*description*)
	      term))       (*id | struct-var*) 
		list),
      pre: term list,      (*preconditions*)
      (*script*)  
      scr: scr
	   };
val e_meth' = {rew_ord' = "e_rew_ord'": rew_ord',
	      erls = e_rls, srls = e_rls, prls = e_rls,
	      calc = [],
	      asm_thm = []: thm' list,
	      asm_rls = []: rls' list,
	      ppc = []: (string * (term * term)) list,
	      pre = []: term list,
	      scr = EmptyScr: scr}:meth';


(** problem-types stored in format for usage in specify  **)
(*25.8.01 ----
val pbltypes = ref ([(e_pblID,[])]:(pblID * ((string * (* field "#Given",..*)
			     (term *   (* description      *)
			      term))    (* id | struct-var  *)
			     list)
		    ) list);*)
type pbt = 
     {thy   : theory,     (* which allows to compile that pbt
			  TODO: search generalized for subthy (ref.p.69*)
      cas   : term option,(*'CAS-command'*)
      prls  : rls,     (* for preds in where_*)
      where_: term list,  (* where - predicates*)
      ppc   : ((string *  (* field "#Given",..*)
		(term *   (* description      *)
		 term))   (* id | struct-var  *)
		   list),
      met   : metID list}; (* methods solving the pbt*)
val e_pbt = {thy=ProtoPure.thy,cas=None,prls=Erls,
	     where_=[],ppc=[],met=[]}:pbt;

datatype ptyp = 
	 Ptyp of string *    (*key within pblID*)
		 pbt list *  (*several pbts with different domIDs
			      TODO: select by subthy (ref.p.69
			      presently only _one_ elem*)
		 ptyp list;
val e_Ptyp = Ptyp ("e_pblID",[e_pbt],[]);

type ptyps = ptyp list;
val ptyps = ref ([e_Ptyp]:ptyps);

(**+ breadth-first search on hierarchy of problem-types +**)

type pblRD = pblID;(*pblID are Reverted _on calling_ the retrieve-funs*)

fun get_py thy d _ [] = 
    raise error ("get_pbt not found: "^(strs2str d))
  | get_py thy d (k::[]) ((Ptyp (k',[py],_))::pys) =
    if k=k' then py
    else get_py thy d ([k]:pblRD) pys
  | get_py thy d (k::ks) ((Ptyp (k',_,pys))::pys') =
    if k=k' then get_py thy d ks pys
    else get_py thy d (k::ks) pys';
(*> ptyps:= 
[Ptyp ("1",[("ptyp 1",([],[]))],
	[Ptyp ("11",[("ptyp 11",([],[]))],
		[])
	 ]),
 Ptyp ("2",[("ptyp 2",([],[]))],
	[Ptyp ("21",[("ptyp 21",([],[]))],
		[])
	 ])
 ];
> get_py SqRoot.thy ["1"] ["1"] (!ptyps);
> get_py SqRoot.thy ["2","21"] ["2","21"] (!ptyps);
         _REVERSE_  .......... !!!!!!!!!!*)

(*TODO: search generalized for subthy*)
fun get_pbt (*thy:theory TODO:subthy*) (pblID:pblID) =
    let val pblRD = rev pblID;
    in get_py  ProtoPure.thy pblID pblRD (!ptyps) end;
(* get_pbt thy ["1"];
   get_pbt thy ["21","2"];
   *)




val methods = 
  ref ([]:(metID * meth') list);


fun get_met metID  = (the (assoc (!methods, metID)))
  handle _ => error ("unknown method: "^(metID2str metID));

